"""Block graph representation of a logical computation."""

from __future__ import annotations

import json
import math
import pathlib
from collections.abc import Iterable, Mapping
from copy import deepcopy
from io import BytesIO
from typing import TYPE_CHECKING, Any, cast

import numpy as np
from networkx import Graph, is_connected
from networkx.utils import graphs_equal

from tqec.computation.correlation import find_correlation_surfaces
from tqec.computation.cube import Cube, CubeKind, Port, YHalfCube, ZXCube, cube_kind_from_string
from tqec.computation.pipe import Pipe, PipeKind
from tqec.utils.enums import Basis
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D, Position3D, SignedDirection3D

if TYPE_CHECKING:
    from tqec.computation.correlation import CorrelationSurface
    from tqec.computation.open_graph import FilledGraph
    from tqec.interop.collada.html_viewer import _ColladaHTMLViewer
    from tqec.interop.pyzx.positioned import PositionedZX


BlockKind = CubeKind | PipeKind


class BlockGraph:
    _NODE_DATA_KEY: str = "tqec_node_data"
    _EDGE_DATA_KEY: str = "tqec_edge_data"

    def __init__(self, name: str = "") -> None:
        """Block graph representation of a logical computation.

        A block graph consists of building blocks that fully define the boundary
        conditions and topological structures of a logical computation. It corresponds
        to the commonly used 3D spacetime diagram representation of a surface code
        logical computation.

        The graph contains two categories of blocks:

        1. :py:class:`~tqec.computation.cube.Cube`: The fundamental building blocks
        of the computation. A cube represents a block of quantum operations within
        a specific spacetime volume. These operations preserve or manipulate the
        quantum information encoded in the logical qubits. Cubes are represented
        as nodes in the graph.

        2. :py:class:`~tqec.computation.pipe.Pipe`: Connects cubes to form the
        topological structure representing the logical computation. A pipe occupies
        no spacetime volume and only replaces the operations within the cubes it
        connects. Pipes are represented as edges in the graph.

        Args:
            name: user-readable name representing the block-graph. Not used internally, but useful
                to identify :class:`.BlockGraph` instances.

        """
        self._name = name
        self._graph: Graph[Position3D] = Graph()
        self._ports: dict[str, Position3D] = {}

    @property
    def name(self) -> str:
        """Name of the graph."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    def num_cubes(self) -> int:
        """Number of cubes (nodes) in the graph, including the ports."""
        return self._graph.number_of_nodes()

    @property
    def num_pipes(self) -> int:
        """Number of pipes (edges) in the graph."""
        return self._graph.number_of_edges()

    @property
    def num_ports(self) -> int:
        """Number of ports in the graph."""
        return len([node for node in self.cubes if node.is_port])

    @property
    def num_half_y_cubes(self) -> int:
        """Number of half Y cubes in the graph."""
        return len([node for node in self.cubes if node.is_y_cube])

    @property
    def ordered_ports(self) -> list[str]:
        """Get the labels of the ports in the alphabetical order."""
        return sorted(self._ports.keys())

    @property
    def cubes(self) -> list[Cube]:
        """The list of cubes (nodes) in the graph."""
        return [data[self._NODE_DATA_KEY] for _, data in self._graph.nodes(data=True)]

    @property
    def pipes(self) -> list[Pipe]:
        """The list of pipes (edges) in the graph."""
        return [data[self._EDGE_DATA_KEY] for _, _, data in self._graph.edges(data=True)]

    @property
    def occupied_positions(self) -> list[Position3D]:
        """Get the positions occupied by the cubes in the graph."""
        return list(self._graph.nodes)

    @property
    def is_open(self) -> bool:
        """Whether the graph is an open graph, i.e. the graph has ports."""
        return bool(self._ports)

    @property
    def spacetime_volume(self) -> float:
        """Return the spacetime volume of the computation.

        A port cube and the pipes have no spacetime volume. A half Y cube has a
        spacetime volume of 0.5. Other cubes have a spacetime volume of 1. The
        spacetime volume of the block graph is the sum of the spacetime volumes
        of all the cubes in the graph.

        Returns:
            The spacetime volume of the computation.

        """
        return self.num_cubes - self.num_ports - self.num_half_y_cubes * 0.5

    def bounding_box_size(self) -> tuple[int, int, int]:
        """Return the size of the bounding box of the computation structure.

        Returns:
            A tuple of three integers representing the width along the X, Y, and Z
            directions, respectively.

        """
        positions = self.occupied_positions
        return (
            max(pos.x for pos in positions) - min(pos.x for pos in positions) + 1,
            max(pos.y for pos in positions) - min(pos.y for pos in positions) + 1,
            max(pos.z for pos in positions) - min(pos.z for pos in positions) + 1,
        )

    @property
    def ports(self) -> dict[str, Position3D]:
        """Mapping from port labels to their positions.

        A port is a virtual node with a unique label that represents the input/output of the
        computation. It should be invisible when visualizing the computation model.

        """
        return dict(self._ports)

    def get_degree(self, position: Position3D) -> int:
        """Get the degree of a node in the graph, i.e. the number of edges incident to it."""
        return self._graph.degree(position)  # type: ignore

    @property
    def leaf_cubes(self) -> list[Cube]:
        """Get the leaf cubes of the graph, i.e. the cubes with degree 1."""
        return [node for node in self.cubes if self.get_degree(node.position) == 1]

    def _check_cube_exists(self, position: Position3D) -> None:
        """Check if a cube exists at the given position."""
        if position not in self:
            raise TQECError(f"No cube at position {position}.")

    def _check_pipe_exists(self, pos1: Position3D, pos2: Position3D) -> None:
        """Check if a pipe exists between the given positions."""
        if not self.has_pipe_between(pos1, pos2):
            raise TQECError(f"No pipe between {pos1} and {pos2}.")

    def add_cube(self, position: Position3D, kind: CubeKind | str, label: str = "") -> Position3D:
        """Add a cube to the graph.

        Args:
            position: The position of the cube.
            kind: The kind of the cube. It can be a :py:class:`~tqec.computation.cube.CubeKind`
                instance or a string representation of the cube kind.
            label: The label of the cube. Default is None.

        Returns:
            The position of the cube added to the graph.

        Raises:
            TQECError: If there is already a cube at the same position, or
                if the cube kind is not recognized, or if the cube is a port and
                there is already a port with the same label in the graph.

        """
        if position in self:
            raise TQECError(f"Cube already exists at position {position}.")
        if isinstance(kind, str):
            kind = cube_kind_from_string(kind)
        if kind == Port() and label in self._ports:
            raise TQECError(f"There is already a port with the same label {label} in the graph.")

        self._graph.add_node(position, **{self._NODE_DATA_KEY: Cube(position, kind, label)})
        if kind == Port():
            self._ports[label] = position
        return position

    def add_pipe(
        self, pos1: Position3D, pos2: Position3D, kind: PipeKind | str | None = None
    ) -> None:
        """Add a pipe to the graph.

        .. note::
            The validity of the pipe WILL NOT be checked when adding it to the graph.
            This allows the user to construct the invalid graph and visualize it for whatever
            purpose. To check the validity of the graph, use the
            :py:meth:`~tqec.computation.block_graph.BlockGraph.validate`.

        Args:
            pos1: The position of one end of the pipe.
            pos2: The position of the other end of the pipe.
            kind: The kind of the pipe connecting the cubes. If None, the kind will be
                automatically determined based on the cubes it connects to make the
                boundary conditions consistent. Default is None.

        Raises:
            TQECError: If any of the positions do not have a cube in the graph, or
                if there is already an pipe between the given positions, or
                if the pipe is not compatible with the cubes it connects.

        """
        u, v = self[pos1], self[pos2]
        if self.has_pipe_between(pos1, pos2):
            raise TQECError("There is already a pipe between the given positions in the graph.")
        if kind is None:
            pipe = Pipe.from_cubes(u, v)
        else:
            if isinstance(kind, str):
                kind = PipeKind.from_str(kind)
            pipe = Pipe(u, v, kind)
        self._graph.add_edge(pos1, pos2, **{self._EDGE_DATA_KEY: pipe})

    def remove_cube(self, position: Position3D) -> None:
        """Remove a cube from the graph, as well as the pipes connected to it.

        Args:
            position: The position of the cube to be removed.

        Raises: TQECError: If there is no cube at the given position.

        """
        self._check_cube_exists(position)
        cube = self[position]
        self._graph.remove_node(position)
        if cube.is_port:
            self._ports.pop(cube.label)

    def remove_pipe(self, pos1: Position3D, pos2: Position3D) -> None:
        """Remove a pipe between two positions.

        Args:
            pos1: The position of one end of the pipe.
            pos2: The position of the other end of the pipe.

        Raises:
            TQECError: If there is no pipe between the given positions.

        """
        self._check_pipe_exists(pos1, pos2)
        self._graph.remove_edge(pos1, pos2)

    def has_pipe_between(self, pos1: Position3D, pos2: Position3D) -> bool:
        """Check if there is a pipe between two positions.

        Args:
            pos1: The first endpoint position.
            pos2: The second endpoint position.

        Returns:
            True if there is an pipe between the two positions, False otherwise.

        """
        return self._graph.has_edge(pos1, pos2)

    def get_pipe(self, pos1: Position3D, pos2: Position3D) -> Pipe:
        """Get the pipe by its endpoint positions.

        Args:
            pos1: The first endpoint position.
            pos2: The second endpoint position.

        Returns:
            The pipe between the two positions.

        Raises:
            TQECError: If there is no pipe between the given positions.

        """
        self._check_pipe_exists(pos1, pos2)
        return cast(Pipe, self._graph.edges[pos1, pos2][self._EDGE_DATA_KEY])

    def pipes_at(self, position: Position3D) -> list[Pipe]:
        """Get the pipes incident to a position."""
        if position not in self:
            raise TQECError(f"No cube at position {position}.")
        return [
            cast(Pipe, data[self._EDGE_DATA_KEY])
            for _, _, data in self._graph.edges(position, data=True)
        ]

    def clone(self) -> BlockGraph:
        """Create a data-independent copy of the graph."""
        graph = BlockGraph(self.name + "_clone")
        graph._graph = deepcopy(self._graph)
        graph._ports = dict(self._ports)
        return graph

    def __contains__(self, position: Position3D) -> bool:
        return position in self._graph

    def __getitem__(self, position: Position3D) -> Cube:
        self._check_cube_exists(position)
        return cast(Cube, self._graph.nodes[position][self._NODE_DATA_KEY])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BlockGraph):
            return False  # pragma: no cover
        return (
            graphs_equal(self._graph, other._graph)  # type: ignore
            and self._ports == other._ports
        )

    def __hash__(self) -> int:
        return hash(self._name)  # pragma: no cover

    def validate(self) -> None:
        """Check the validity of the block graph to represent a logical computation.

        Refer to the Fig.9 in arXiv:2404.18369. Currently, we ignore the b) and e),
        only check the following conditions:

        - **No fanout:** ports can only have one pipe connected to them.
        - **Time-like Y:** Y Half Cubes can only have time-like pipes connected to them.
        - **No 3D corner:** a cube cannot have pipes in all three directions.
        - **Match color at passthrough:** two pipes in a "pass-through" should have the same
          color orientation.
        - **Match color at turn:** two pipes in a "turn" should have the matching colors on
          faces that are touching.

        Raises:
            TQECError: If the above conditions are not satisfied.

        """
        for cube in self.cubes:
            self._validate_locally_at_cube(cube)

    def _validate_locally_at_cube(self, cube: Cube) -> None:
        """Check the validity of the block structures locally at a cube."""
        pipes = self.pipes_at(cube.position)
        # no fanout at ports
        if cube.is_port:
            if len(pipes) != 1:
                raise TQECError(
                    f"Port at {cube.position} does not have exactly one pipe connected."
                )
            return
        # time-like Y
        if cube.is_y_cube:
            if len(pipes) != 1:
                raise TQECError(
                    f"Y Half Cube at {cube.position} does not have exactly one pipe connected."
                )
            if not pipes[0].direction == Direction3D.Z:
                raise TQECError(f"Y Half Cube at {cube.position} has non-timelike pipes connected.")
            return

        assert isinstance(cube.kind, ZXCube)
        # Check the color matching conditions
        pipes_by_direction: dict[Direction3D, list[Pipe]] = {}
        for pipe in pipes:
            pipes_by_direction.setdefault(pipe.direction, []).append(pipe)
        for direction in Direction3D.all_directions():
            # the pair of faces are shadowed in the direction
            # we do not care about the colors of shadowed faces
            if len(pipes_by_direction.get(direction, [])) == 2:
                continue
            # faces at the same plane should have the same color
            cube_color = cube.kind.get_basis_along(direction)
            for ortho_dir in direction.orthogonal_directions:
                for pipe in pipes_by_direction.get(ortho_dir, []):
                    pipe_color = pipe.kind.get_basis_along(direction, pipe.at_head(cube.position))
                    if pipe_color != cube_color:
                        raise TQECError(f"Cube {cube} has mismatched colors with pipe {pipe}.")

    def to_zx_graph(self) -> PositionedZX:
        """Convert the block graph to a positioned PyZX graph.

        Returns:
            A :py:class:`~tqec.interop.pyzx.positioned.PositionedZX` object
            converted from the block graph.

        """
        # Needs to be imported here to avoid pulling pyzx when importing this module.
        from tqec.interop.pyzx.positioned import PositionedZX  # noqa: PLC0415

        return PositionedZX.from_block_graph(self)

    def to_dae_file(
        self,
        file_path: str | pathlib.Path,
        pipe_length: float = 2.0,
        pop_faces_at_directions: Iterable[SignedDirection3D | str] = (),
        show_correlation_surface: CorrelationSurface | None = None,
    ) -> None:
        """Write the block graph to a Collada DAE file.

        Args:
            file_path: The output file path.
            pipe_length: The length of the pipes. Default is 2.0.
            pop_faces_at_directions: Remove the faces at the given directions for all the blocks.
                This is useful for visualizing the internal structure of the blocks.
                Default is None.
            show_correlation_surface: The correlation surface to show in the block graph.
                Default is None.

        """
        # Needs to be imported here to avoid pulling collada when importing this module.
        from tqec.interop.collada.read_write import write_block_graph_to_dae_file  # noqa: PLC0415

        write_block_graph_to_dae_file(
            self,
            file_path,
            pipe_length,
            pop_faces_at_directions,
            show_correlation_surface,
        )

    @staticmethod
    def from_dae_file(filename: str | pathlib.Path, graph_name: str = "") -> BlockGraph:
        """Construct a block graph from a COLLADA DAE file.

        Args:
            filename: The input ``.dae`` file path.
            graph_name: The name of the block graph. Default is an empty string.

        Returns:
            The :py:class:`~tqec.computation.block_graph.BlockGraph` object
            constructed from the DAE file.

        """
        # Needs to be imported here to avoid pulling collada when importing this module.
        from tqec.interop.collada.read_write import read_block_graph_from_dae_file  # noqa: PLC0415

        return read_block_graph_from_dae_file(filename, graph_name)

    def view_as_html(
        self,
        write_html_filepath: str | pathlib.Path | None = None,
        pipe_length: float = 2.0,
        pop_faces_at_directions: Iterable[SignedDirection3D | str] = (),
        show_correlation_surface: CorrelationSurface | None = None,
    ) -> _ColladaHTMLViewer:
        """View COLLADA model in html with the help of ``three.js``.

        This can display a COLLADA model interactively in IPython compatible environments.

        Args:
            write_html_filepath: The output html file path to write the generated html content
                if provided. Default is None.
            pipe_length: The length of the pipes. Default is 2.0.
            pop_faces_at_directions: Remove the faces at the given directions for all the blocks.
                This is useful for visualizing the internal structure of the blocks.
                Default is None.
            show_correlation_surface: The correlation surface to show in the block graph.
                Default is None.

        Returns:
            A helper class to display the 3D model, which implements the ``_repr_html_`` method and
            can be directly displayed in IPython compatible environments.

        """
        # Needs to be imported here to avoid pulling collada when importing this module.
        from tqec.interop.collada.html_viewer import display_collada_model  # noqa: PLC0415
        from tqec.interop.collada.read_write import write_block_graph_to_dae_file  # noqa: PLC0415

        bytes_buffer = BytesIO()
        write_block_graph_to_dae_file(
            self,
            bytes_buffer,
            pipe_length,
            pop_faces_at_directions,
            show_correlation_surface,
        )
        return display_collada_model(
            filepath_or_bytes=bytes_buffer.getvalue(),
            write_html_filepath=write_html_filepath,
        )

    def shift_by(self, dx: int = 0, dy: int = 0, dz: int = 0) -> BlockGraph:
        """Shift a copy of ``self`` by the given offset in the x, y, z directions and return it.

        Args:
            dx: The offset in the x direction.
            dy: The offset in the y direction.
            dz: The offset in the z direction.

        Returns:
            A new graph with the shifted positions. The new graph will share no data
            with the original graph.

        """
        new_graph = BlockGraph()
        for cube in self.cubes:
            new_graph.add_cube(cube.position.shift_by(dx=dx, dy=dy, dz=dz), cube.kind, cube.label)
        for pipe in self.pipes:
            u, v = pipe.u, pipe.v
            new_graph.add_pipe(
                u.position.shift_by(dx=dx, dy=dy, dz=dz),
                v.position.shift_by(dx=dx, dy=dy, dz=dz),
                pipe.kind,
            )
        return new_graph

    def find_correlation_surfaces(self) -> list[CorrelationSurface]:
        """Find the correlation surfaces in the block graph.

        Returns:
            The list of correlation surfaces.

        """
        return find_correlation_surfaces(self.to_zx_graph().g)

    def fill_ports(self, fill: Mapping[str, CubeKind] | CubeKind) -> None:
        """Fill the ports at specified positions with cubes of the given kind.

        Args:
            fill: A mapping from the label of the ports to the cube kind to fill.
                If a single kind is given, all the ports will be filled with the
                same kind.

        Raises:
            TQECError: if there is no port with the given label.

        """
        if isinstance(fill, CubeKind):
            fill = {label: fill for label in self._ports}
        for label, kind in fill.items():
            if label not in self._ports:
                raise TQECError(f"There is no port with label {label}.")
            pos = self._ports[label]
            fill_node = Cube(pos, kind)
            # Overwrite the node at the port position
            self._graph.add_node(pos, **{self._NODE_DATA_KEY: fill_node})
            for pipe in self.pipes_at(pos):
                self._graph.remove_edge(pipe.u.position, pipe.v.position)
                other = pipe.u if pipe.v.position == pos else pipe.v
                self._graph.add_edge(
                    other.position,
                    pos,
                    **{self._EDGE_DATA_KEY: Pipe(other, fill_node, pipe.kind)},
                )
            # Delete the port label
            self._ports.pop(label)

    def fill_ports_for_minimal_simulation(self) -> list[FilledGraph]:
        """Fill the ports of the provided ``graph`` to minimize the number of simulation runs.

        Given a block graph with open ports, fill in the ports with the appropriate cubes that
        will minimize the number of simulation runs needed for the complete logical observable set.

        Returns:
            A list of :class:`~tqec.computation.open_graph.FilledGraph` instances, each
            containing a block graph with all ports filled and a set of correlation
            surfaces that can be used as logical observables for the simulation on that
            block graph.

        """
        # Needs to be imported here to avoid pulling pyzx when importing this module.
        from tqec.computation.open_graph import fill_ports_for_minimal_simulation  # noqa: PLC0415

        return fill_ports_for_minimal_simulation(self)

    def compose(self, other: BlockGraph, self_port: str, other_port: str) -> BlockGraph:
        """Compose the current graph with another graph.

        The other graph will be shifted to match the ports in the current graph
        and the two graphs will be composed at the corresponding ports.

        The two ports provided to this method will serve as an "anchor", and other
        overlapping ports will be detected and glued automatically.

        Args:
            other: The other graph to be composed with the current graph.
            self_port: The label of the port to be connected in the current graph.
            other_port: The label of the port to be connected in the other graph.

        Returns:
            A new graph that is the composition of the current graph with the other graph.

        Raises:
            TQECError: If the ports are not in the graphs, or if the two graphs
                cannot be composed because overlapping spacetime extents or incompatible
                cube kinds.

        """
        if self_port not in self.ports:
            raise TQECError(f"Port {self_port} is not in the current graph.")
        if other_port not in other.ports:
            raise TQECError(f"Port {other_port} is not in the other graph.")

        p1, p2 = self.ports[self_port], other.ports[other_port]
        shift = (p1.x - p2.x, p1.y - p2.y, p1.z - p2.z)
        shifted_g = other.shift_by(*shift)

        ports_need_fill: dict[str, Position3D] = {}
        # Check if there is overlapping spacetime extent
        for cube in self.cubes:
            pos = cube.position
            if pos in shifted_g:
                if cube.is_port and shifted_g[pos].is_port:
                    ports_need_fill[cube.label] = pos
                    continue
                raise TQECError(
                    f"Cube at position {cube.position} is overlapping between the two graphs."
                )
        composed_g = self.clone()
        # Resolve the cube kinds at the ports
        for label, port in ports_need_fill.items():
            bases: list[Basis] = []
            # port is guaranteed to only have one pipe connected
            pipe1 = self.pipes_at(port)[0]
            pipe2 = shifted_g.pipes_at(port)[0]
            for d in Direction3D.all_directions():
                b1 = pipe1.kind.get_basis_along(d, pipe1.at_head(port))
                b2 = pipe2.kind.get_basis_along(d, pipe2.at_head(port))
                if b1 is not None and b2 is not None and b1 != b2:
                    raise TQECError(
                        f"Port at {port} cannot be filled with a cube that has "
                        "valid boundary conditions."
                    )
                # choose Z basis boundary for the walls that can have arbitrary boundary
                bases.append(b1 or b2 or Basis.Z)
            cube_kind = ZXCube(*bases)
            composed_g.fill_ports({label: cube_kind})
        # Compose the graphs
        for cube in shifted_g.cubes:
            # Connecting ports have been filled
            if cube.position in composed_g:
                continue
            composed_g.add_cube(cube.position, cube.kind, cube.label)
        for pipe in shifted_g.pipes:
            u, v = pipe.u.position, pipe.v.position
            composed_g.add_pipe(u, v, pipe.kind)
        composed_g.ports.update({s: p for s, p in shifted_g.ports.items() if composed_g[p].is_port})
        composed_g.name = f"{self.name}_composed_with_{other.name}"
        return composed_g

    def is_single_connected(self) -> bool:
        """Check if the graph is single connected.

        A block graph is single-connected if there is only one connected component in the graph.
        """
        return bool(is_connected(self._graph))  # type: ignore[invalid-argument-type]

    def rotate(
        self,
        rotation_axis: Direction3D,
        counterclockwise: bool = True,
        num_90_degree_rotation: int = 1,
    ) -> BlockGraph:
        """Rotate a copy of ``self`` around an axis by a multiple of 90 degrees and return it.

        Args:
            rotation_axis: The axis to rotate around.
            counterclockwise: Whether to rotate counterclockwise. Default is True.
            num_90_degree_rotation: The number of 90-degree rotations. Default is 1.

        Returns:
            A new graph with the rotated positions. The new graph will share no data
            with the original graph.

        """
        # Needs to be imported here to avoid pulling scipy when importing this module.
        from tqec.utils.rotations import (  # noqa: PLC0415
            get_rotation_matrix,
            rotate_block_kind_by_matrix,
            rotate_position_by_matrix,
        )

        rotated = BlockGraph(self.name + "_rotated")
        rotation_matrix = get_rotation_matrix(
            rotation_axis, counterclockwise, num_90_degree_rotation * math.pi / 2
        )
        pos_map: dict[Position3D, Position3D] = {}
        for cube in self.cubes:
            rotated_kind = rotate_block_kind_by_matrix(cube.kind, rotation_matrix)
            rotated_pos = rotate_position_by_matrix(cube.position, rotation_matrix)
            rotated.add_cube(rotated_pos, cast(CubeKind, rotated_kind), cube.label)
            pos_map[cube.position] = rotated_pos

        for pipe in self.pipes:
            rotated_kind = rotate_block_kind_by_matrix(pipe.kind, rotation_matrix)
            rotated.add_pipe(
                pos_map[pipe.u.position],
                pos_map[pipe.v.position],
                cast(PipeKind, rotated_kind),
            )
        return rotated

    def fix_shadowed_faces(self) -> BlockGraph:
        """Fix the basis of those shadowed faces of the cubes in the graph.

        A pair of face can be shadowed if the cube is connected to two pipes
        in the same direction. Though these faces are not visible in the 3D
        visualization, they are still identified by the cube kind and can
        affect the circuit compilation.

        The basis of the cube is enforced to match the pipes connected to it.
        If there is unmatched shadowed faces, we try to fix their basis. This
        will leave some freedom in constructing the model in SketchUp and make
        everyones' life easier.

        Additionally, for a spatial pass-through, i.e. a cube only connected to
        two pipes in the same direction, we will fix the cube kind to not be a
        spatial cube.

        Note that the fixed graph is still not guaranteed to be valid as some
        other conditions may be violated. You still need to call
        :py:meth:`~tqec.computation.block_graph.BlockGraph.validate` to check
        the validity of the graph.

        Returns:
            A new graph with the shadowed faces fixed.

        """
        fixed_cubes: dict[Cube, Cube] = {}
        for cube in self.cubes:
            if not isinstance(cube.kind, ZXCube):
                continue
            # Group connected pipes by direction
            pipes_by_direction: dict[Direction3D, list[Pipe]] = {}
            for pipe in self.pipes_at(cube.position):
                pipes_by_direction.setdefault(pipe.direction, []).append(pipe)
            # No need to handle the case `len(pipes_by_direction) == 0` as there
            # is no pipes connected to the cube.
            # No need to handle the case `len(pipes_by_direction) == 3` as it's
            # a 3D corner that cannot be a valid structure.
            # Spatial pass-through, ensure that the cube is not a spatial cube
            if len(pipes_by_direction) in [0, 3]:
                continue
            shadowed_directions = {d for d, ps in pipes_by_direction.items() if len(ps) == 2}
            if not shadowed_directions:
                continue
            new_kind = cube.kind
            for shadowed_direction in shadowed_directions:
                # Spatial pass-through, ensure that the cube is not a spatial cube
                if (
                    len(pipes_by_direction) == 1
                    and shadowed_direction != Direction3D.Z
                    and cube.is_spatial
                ):
                    kind = cube.kind
                    assert isinstance(kind, ZXCube)
                    basis = kind.get_basis_along(shadowed_direction)
                    new_kind = kind.with_basis_along(shadowed_direction, basis.flipped())
                    new_cube = Cube(cube.position, new_kind, cube.label)
                    fixed_cubes[cube] = new_cube
                # T-shape or X-shape connections, can be either in space or time
                # Ensure the shadowed faces match the pipe faces that are in the
                # same plane.
                elif len(pipes_by_direction) == 2:
                    other_direction = next(d for d in pipes_by_direction if d != shadowed_direction)
                    need_match_pipe = next(iter(pipes_by_direction[other_direction]))
                    pipe_basis = need_match_pipe.kind.get_basis_along(
                        shadowed_direction, need_match_pipe.at_head(cube.position)
                    )
                    assert pipe_basis is not None
                    new_kind = new_kind.with_basis_along(shadowed_direction, pipe_basis)

            if new_kind != cube.kind:
                new_cube = Cube(cube.position, new_kind, cube.label)
                fixed_cubes[cube] = new_cube
        new_graph = BlockGraph(self.name)
        for cube in self.cubes:
            new_cube = fixed_cubes.get(cube, cube)
            new_graph.add_cube(cube.position, new_cube.kind, new_cube.label)
        for pipe in self.pipes:
            new_graph.add_pipe(pipe.u.position, pipe.v.position, pipe.kind)
        return new_graph

    def get_cubes_by_label(self, label: str) -> list[Cube]:
        """Find cubes with the specified label in the BlockGraph.

        Args:
            label: The label of the cubes.

        Returns:
            The cube instances that have the specified label.

        """
        return [cube for cube in self.cubes if cube.label == label]

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the block graph."""
        return {
            "name": self.name,
            "cubes": [cube.to_dict() for cube in self.cubes],
            "pipes": [pipe.to_dict() for pipe in self.pipes],
            "ports": {label: pos.as_tuple() for label, pos in self.ports.items()},
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> BlockGraph:
        """Construct a block graph from a dictionary representation."""
        graph = BlockGraph(data["name"])
        for cube in data["cubes"]:
            graph.add_cube(
                position=Position3D(*cube["position"]),
                kind=cube["kind"],
                label=cube["label"],
            )
        for pipe in data["pipes"]:
            graph.add_pipe(
                pos1=Position3D(*pipe["u"]),
                pos2=Position3D(*pipe["v"]),
                kind=pipe["kind"],
            )
        return graph

    def to_json(
        self,
        file_path: str | pathlib.Path | None = None,
        *,
        indent: int | None = 2,
    ) -> str | None:
        """Serialize the block graph to a JSON string or write it to a file.

        Args:
            file_path: The output file path. If None, the JSON string will be returned.
            indent: The indentation level for pretty printing, passed to
                `json.dumps`. Default is 2.

        Returns:
            The JSON string representation of the block graph if `file_path` is None,
            otherwise None.

        """
        # Get dictionary from blockgraph
        obj_dict = self.to_dict()

        # Add transform matrix to keep schema symmetric
        for cube in obj_dict["cubes"]:
            cube["transform"] = np.eye(3, dtype=int).tolist()

        for pipe in obj_dict["pipes"]:
            pipe["transform"] = np.eye(3, dtype=int).tolist()

        # Return or write
        if file_path is None:
            return json.dumps(obj_dict, indent=indent)
        with open(file_path, "w") as fp:
            json.dump(obj_dict, fp, indent=indent)
            return None

    @staticmethod
    def from_json(filename: str | pathlib.Path, graph_name: str = "") -> BlockGraph:
        """Construct a block graph from a JSON file.

        Args:
            filename: The input ``.json`` file path.
            graph_name: The name of the block graph. Default is an empty string.

        Returns:
            The :py:class:`~tqec.computation.block_graph.BlockGraph` object
            constructed from the DAE file.

        """
        # Needs to be imported here to avoid pulling collada when importing this module.
        from tqec.interop.collada.read_write import read_block_graph_from_json  # noqa: PLC0415

        return read_block_graph_from_json(filename, graph_name)

    def relabel_cubes(self, label_mapping: Mapping[Position3D | str, str]) -> None:
        """Relabel cubes in the block graph.

        This method updates the labels of cubes in the graph, based on a mapping
        from either a cube position or its existing label to a new label.

        Args:
            label_mapping: A mapping from either Position3D or current cube label (str)
                to the new label to assign.

        Raises:
            TQECError: If a cube is not found for the given key, if a port label
                is reused, or if the new label conflicts with existing port labels.

        """
        port_labels = {cube.label for cube in self.cubes if cube.is_port}
        assigned_new_labels: set[str] = set()

        for key, new_label in label_mapping.items():
            if not new_label:
                raise TQECError("New label must be non-empty.")

            if isinstance(key, Position3D):
                matching_cubes = [cube for cube in self.cubes if cube.position == key]
            elif isinstance(key, str):
                matching_cubes = [cube for cube in self.cubes if cube.label == key]
            else:
                raise TQECError(f"Invalid identifier '{key}'. Must be Position3D or str.")

            if not matching_cubes:
                raise TQECError(f"No cube found for identifier '{key}'.")

            is_port_relabel = any(cube.is_port for cube in matching_cubes)

            if is_port_relabel:
                if new_label in port_labels:
                    raise TQECError(
                        f"Port label '{new_label}' is already assigned to another port."
                    )
                if new_label in assigned_new_labels:
                    raise TQECError(f"Port label '{new_label}' is reused multiple times.")
                assigned_new_labels.add(new_label)
                self._ports[new_label] = self._ports.pop(matching_cubes[0].label)
            elif new_label in port_labels:
                raise TQECError(
                    f"The label '{new_label}' belongs to a port and "
                    "cannot be reused by a non-port cube."
                )

            for cube in matching_cubes:
                updated_cube = Cube(position=cube.position, kind=cube.kind, label=new_label)
                self._graph.add_node(cube.position, **{self._NODE_DATA_KEY: updated_cube})


def block_kind_from_str(string: str) -> BlockKind:
    """Parse a block kind from a string."""
    string = string.upper()
    if "O" in string:
        return PipeKind.from_str(string)
    elif string == "Y":
        return YHalfCube()
    else:
        return ZXCube.from_str(string)



from __future__ import annotations

import warnings
from collections.abc import Mapping, Sequence
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any

import stim
from typing_extensions import override

from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.compile.detectors.database import CURRENT_DATABASE_VERSION, DetectorDatabase
from tqec.compile.observables.abstract_observable import AbstractObservable
from tqec.compile.observables.builder import ObservableBuilder
from tqec.compile.tree.annotations import LayerTreeAnnotations, Polygon
from tqec.compile.tree.annotators.circuit import AnnotateCircuitOnLayerNode
from tqec.compile.tree.annotators.detectors import AnnotateDetectorsOnLayerNode
from tqec.compile.tree.annotators.observables import annotate_observable
from tqec.compile.tree.annotators.polygons import AnnotatePolygonOnLayerNode
from tqec.compile.tree.node import LayerNode, NodeWalker
from tqec.post_processing.shift import shift_to_only_positive
from tqec.utils.exceptions import TQECError, TQECWarning
from tqec.utils.paths import DEFAULT_DETECTOR_DATABASE_PATH


class QubitLister(NodeWalker):
    def __init__(self, k: int):
        """Keep in memory all the qubits used by the nodes explored.

        Args:
            k: scaling factor used to explore the quantum circuits.

        """
        super().__init__()
        self._k = k
        self._seen_qubits: set[GridQubit] = set()

    @override
    def visit_node(self, node: LayerNode) -> None:
        if not node.is_leaf:
            return
        annotations = node.get_annotations(self._k)
        if annotations.circuit is None:
            raise TQECError("Cannot list qubits without the circuit annotation.")
        self._seen_qubits |= annotations.circuit.qubits

    @property
    def seen_qubits(self) -> set[GridQubit]:
        """Return all the qubits seen when exploring."""
        return self._seen_qubits


class LayerTree:
    def __init__(
        self,
        root: SequencedLayers,
        observable_builder: ObservableBuilder,
        abstract_observables: list[AbstractObservable] | None = None,
        annotations: Mapping[int, LayerTreeAnnotations] | None = None,
    ):
        """Represent a computation as a tree.

        Note:
            It is expected that the root is a
            :class:`~tqec.compile.blocks.layer.composed.sequenced.SequencedLayers`
            with each of its children representing the computation on a timespan
            equivalent to the timespan of a cube.

        Args:
            root: root node of the tree.
            abstract_observables: a list of abstract observables to be compiled into
                observables. If set to ``None``, no observables will be compiled
                into the circuit.
            annotations: a mapping from positive integers representing the value
                of ``k``, the scaling factor, to annotations computed for that
                value of ``k``.
            observable_builder: the style of the surface code patch.

        """
        self._root = LayerNode(root)
        self._abstract_observables = abstract_observables or []
        self._annotations = dict(annotations) if annotations is not None else {}
        self._observable_builder = observable_builder

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of ``self``."""
        return {  # pragma: no cover
            "root": self._root.to_dict(),
            "abstract_observables": self._abstract_observables,
            "annotations": {k: annotation.to_dict() for k, annotation in self._annotations.items()},
        }

    def _annotate_circuits(self, k: int, reschedule_measurements: bool = True) -> None:
        self._root.walk(
            AnnotateCircuitOnLayerNode(k, reschedule_measurements=reschedule_measurements)
        )

    def _annotate_qubit_map(self, k: int) -> None:
        self._get_annotation(k).qubit_map = self._get_global_qubit_map(k)

    def _get_global_qubit_map(self, k: int) -> QubitMap:
        qubit_lister = QubitLister(k)
        self._root.walk(qubit_lister)
        return QubitMap.from_qubits(sorted(qubit_lister.seen_qubits))

    def _annotate_observables(self, k: int) -> None:
        for obs_idx, observable in enumerate(self._abstract_observables):
            annotate_observable(self._root, k, observable, obs_idx, self._observable_builder)

    def _annotate_detectors(
        self,
        k: int,
        manhattan_radius: int = 2,
        detector_database: DetectorDatabase | None = None,
        database_path: Path | None = DEFAULT_DETECTOR_DATABASE_PATH,
        only_use_database: bool = False,
        lookback: int = 2,
        parallel_process_count: int = 1,
    ) -> None:
        if manhattan_radius <= 0:
            return  # pragma: no cover
        self._root.walk(
            AnnotateDetectorsOnLayerNode(
                k,
                manhattan_radius,
                detector_database,
                only_use_database,
                lookback,
                parallel_process_count,
            )
        )
        # The database will have been updated inside the above function, and here at
        # the end of the computation we save it to file.
        if detector_database is not None and database_path is not None:
            detector_database.to_file(database_path)

    def _annotate_polygons(
        self,
        k: int,
    ) -> None:
        self._root.walk(AnnotatePolygonOnLayerNode(k))  # pragma: no cover

    def generate_crumble_url(
        self,
        k: int,
        manhattan_radius: int = 2,
        detector_database: DetectorDatabase | None = None,
        lookback: int = 2,
        add_polygons: bool = True,
        shift_to_positive: bool = True,
    ) -> str:
        """Generate the Crumble URL of the quantum circuit representing ``self``.

        This method first annotates the tree according to the provided arguments
        and then use these annotations to generate the final quantum circuit and
        convert it to a Crumble URL.

        Args:
            k: scaling factor.
            manhattan_radius: Parameter for the automatic computation of detectors.
                Should be large enough so that flows canceling each other to
                form a detector are strictly contained in plaquettes that are at
                most at a distance of ``manhattan_radius`` from the central
                plaquette. Detector computation runtime grows with this parameter,
                so you should try to keep it to its minimum. A value too low might
                produce invalid detectors.
            detector_database: existing database of detectors that is used to
                avoid computing detectors if the database already contains them.
                Default to `None` which result in not using any kind of database
                and unconditionally performing the detector computation.
            lookback: number of QEC rounds to consider to try to find detectors.
                Including more rounds increases computation time.
            add_polygons: whether to include polygons in the Crumble URL. If
                ``True``, the polygons representing the stabilizers will be generated
                based on the RPNG information of underlying plaquettes and add
                to the Crumble URL.
            shift_to_positive: if ``True``, the resulting circuit is shifted such
                that only qubits with positive coordinates are used. Else, the
                circuit is left as is.

        Returns:
            a string representing the Crumble URL of the quantum circuit.

        """
        if not add_polygons:
            circuit = self.generate_circuit(
                k,
                include_qubit_coords=True,
                manhattan_radius=manhattan_radius,
                detector_database=detector_database,
                lookback=lookback,
            )
            return str(circuit.to_crumble_url())
        self._generate_annotations(k, manhattan_radius, detector_database, lookback=lookback)
        self._annotate_polygons(k)
        annotations = self._get_annotation(k)
        qubit_map = annotations.qubit_map
        assert qubit_map is not None
        circuits_with_polygons = self._root.generate_circuits_with_potential_polygons(
            k, qubit_map, add_polygons=True
        )
        qubit_map_circuit = qubit_map.to_circuit()
        if shift_to_positive:
            qubit_map_circuit = shift_to_only_positive(qubit_map_circuit)
        crumble_url: str = qubit_map_circuit.to_crumble_url() + ";"
        last_polygons: set[Polygon] = set()
        for item in circuits_with_polygons:
            if isinstance(item, stim.Circuit):
                circuit_crumble_url = item.to_crumble_url().replace(
                    "https://algassert.com/crumble#circuit=", ""
                )
                crumble_url += circuit_crumble_url
                crumble_url += ";"
            else:
                polygons = set(item)
                if polygons == last_polygons:
                    continue
                crumble_url += "".join(polygon.to_crumble_url_string(qubit_map) for polygon in item)
                last_polygons = polygons
        return crumble_url

    def _generate_annotations(
        self,
        k: int,
        manhattan_radius: int = 2,
        detector_database: DetectorDatabase | None = None,
        database_path: Path | None = None,
        only_use_database: bool = False,
        lookback: int = 2,
        parallel_process_count: int = 1,
        reschedule_measurements: bool = True,
    ) -> None:
        """Annotate the tree with circuits, qubit maps, detectors and observables."""
        # If already annotated, no need to re-annotate.
        if k in self._annotations:
            return  # pragma: no cover
        # Else, perform all the needed computations.
        self._annotate_circuits(k, reschedule_measurements=reschedule_measurements)
        self._annotate_qubit_map(k)
        # This method will also update the detector_database and save it to disk at database_path.
        self._annotate_detectors(
            k,
            manhattan_radius,
            detector_database,
            database_path,
            only_use_database,
            lookback,
            parallel_process_count,
        )
        self._annotate_observables(k)

    def generate_circuit(
        self,
        k: int,
        include_qubit_coords: bool = True,
        manhattan_radius: int = 2,
        detector_database: DetectorDatabase | None = None,
        database_path: str | Path | None = DEFAULT_DETECTOR_DATABASE_PATH,
        do_not_use_database: bool = False,
        only_use_database: bool = False,
        lookback: int = 2,
        reschedule_measurements: bool = True,
    ) -> stim.Circuit:
        """Generate the quantum circuit representing ``self``.

        This method first annotates the tree according to the provided arguments
        and then use these annotations to generate the final quantum circuit.

        Args:
            k: scaling factor.
            include_qubit_coords: whether to include ``QUBIT_COORDS`` annotations
                in the returned quantum circuit or not. Default to ``True``.
            manhattan_radius: Parameter for the automatic computation of detectors.
                Should be large enough so that flows canceling each other to
                form a detector are strictly contained in plaquettes that are at
                most at a distance of ``manhattan_radius`` from the central
                plaquette. Detector computation runtime grows with this parameter,
                so you should try to keep it to its minimum. A value too low might
                produce invalid detectors.
            detector_database: an instance to retrieve from / store in detectors
                that are computed as part of the circuit generation. If not given,
                the detectors are retrieved from/stored in the provided
                ``database_path``.
            database_path: specify where to save to after the calculation.
                This defaults to :data:`.DEFAULT_DETECTOR_DATABASE_PATH` if
                not specified. If detector_database is not passed in, the code attempts to
                retrieve the database from this location.
            do_not_use_database: if ``True``, even the default database will not be used.
            only_use_database: if ``True``, only detectors from the database
                will be used. An error will be raised if a situation that is not
                registered in the database is encountered.
            lookback: number of QEC rounds to consider to try to find detectors.
                Including more rounds increases computation time.
            reschedule_measurements: whether to reschedule measurements in a ``LayoutLayer``
                to be in the same moment. Since each plaquette may have its own measurement
                schedule, setting this may be necessary for hardware that requires
                measurements to be synchronous.

        Returns:
            a ``stim.Circuit`` instance implementing the computation described
            by ``self``.

        """
        db_path_input = DEFAULT_DETECTOR_DATABASE_PATH
        if not do_not_use_database:
            # First, before we start any computations, decide which detector database to use.
            if isinstance(database_path, str):
                db_path_input = Path(database_path)
            else:
                db_path_input = database_path
            # We need to know for later if the user explicitly provided a database or
            # not to decide if we should warn or raise.
            user_defined = (
                detector_database is not None or database_path != DEFAULT_DETECTOR_DATABASE_PATH
            )
            # If the user has passed a database in, use that, otherwise:
            if detector_database is None:  # Nothing passed in,
                if (
                    db_path_input is not None and db_path_input.exists()
                ):  # look for an existing database at the path.
                    detector_database = DetectorDatabase.from_file(db_path_input)
                else:  # if there is no existing database, create one.
                    detector_database = DetectorDatabase()
            loaded_version = detector_database.version
            current_version = CURRENT_DATABASE_VERSION
            if loaded_version != current_version:
                if user_defined:
                    raise TQECError(
                        f"The detector database on disk you have specified is incompatible with"
                        f" the version in the TQEC code you are running. The version of the disk"
                        f" database is {loaded_version}, while the version in the TQEC code is "
                        f"{current_version}."
                    )
                else:  # ie using the default
                    warnings.warn(
                        f"The default detector database that you have saved on your system is out "
                        f"of date (version {loaded_version}). The version in the TQEC code you are "
                        f"running is newer (version {current_version}). The database will be "
                        "regenerated.",
                        TQECWarning,
                    )
                    detector_database = DetectorDatabase()
        else:
            detector_database = None

        # Enable parallel processing only if the detector database is empty or None,
        # as current parallelization is effective only in this case.
        # If we later support efficient parallelism with a populated database,
        # we will expose the parallel_count parameter to users.
        parallel_process_count = (
            cpu_count() // 2 + 1
            if (detector_database is None or len(detector_database) == 0)
            else 1
        )

        self._generate_annotations(
            k,
            manhattan_radius,
            detector_database=detector_database,
            database_path=db_path_input,
            only_use_database=only_use_database,
            lookback=lookback,
            parallel_process_count=parallel_process_count,
            reschedule_measurements=reschedule_measurements,
        )
        annotations = self._get_annotation(k)
        assert annotations.qubit_map is not None

        circuit = stim.Circuit()
        if include_qubit_coords:
            circuit += annotations.qubit_map.to_circuit()
        circuit += self._root.generate_circuit(k, annotations.qubit_map)
        return circuit

    def _get_annotation(self, k: int) -> LayerTreeAnnotations:
        return self._annotations.setdefault(k, LayerTreeAnnotations())

    def layers_to_svg(
        self,
        k: int,
        errors: Sequence[stim.ExplainedError] = tuple(),
        show_observable: int | None = None,
    ) -> list[str]:
        """Visualize the layers as a list of SVG strings.

        Args:
            k: scaling factor.
            errors: a sequence of errors to be drawn on the layers. Each error
                is visualised with a cross. The cross colour follows the XYZ=RGB
                convention, and the moment index at which the error takes place
                is written above the cross (an error is always scheduled at the
                end of the moment, so any operation applied at the same moment
                is applied before the error).
            show_observable: the index of the observable to be drawn on the layers.
                If set to ``None``, no observable will be shown. If set to an
                integer, the observable with that index will be shown. The
                observable is represented as the set of included measurements.
                A yellow star on the plaquette vertex indicates a data qubit
                readout, while a star on the plaquette face indicates a
                stabilizer measurements.

        Returns:
            a list of SVG strings representing the layers of the tree.

        """
        # Warning explicitly disabled because this intended and the only way to
        # avoid the costly svg import.
        from tqec.visualisation.computation.tree import LayerVisualiser  # noqa: PLC0415

        if show_observable is not None and show_observable >= len(self._abstract_observables):
            raise TQECError(
                f"{show_observable:=} is out of range for the number of "
                f"abstract observables ({len(self._abstract_observables)})."
            )
        annotations = self._annotations.get(k, LayerTreeAnnotations())
        tl, br = (
            annotations.qubit_map.qubit_bounds()
            if annotations.qubit_map is not None
            else (None, None)
        )
        # Note: if the top-left and bottom-right qubits are not None, we just computed them from
        # the resulting circuit. We want to stick to the regular plaquette grid, and that might not
        # be the case here because boundary plaquettes might not use the extremal data-qubits.
        # As data-qubits are located on odd coordinates, we just ensure that the boundary
        # coordinates are odd, flooring or ceiling to the closest odd coordinates depending on the
        # boundary to make sure we extended the viewport (and not reduce it).
        if tl is not None:
            tl = GridQubit(tl.x - 1 if tl.x % 2 == 0 else tl.x, tl.y - 1 if tl.y % 2 == 0 else tl.y)
        if br is not None:
            br = GridQubit(br.x + 1 if br.x % 2 == 0 else br.x, br.y + 1 if br.y % 2 == 0 else br.y)
        visualiser = LayerVisualiser(
            k, errors, show_observable, top_left_qubit=tl, bottom_right_qubit=br
        )
        self._root.walk(visualiser)
        return visualiser.visualisations



"""Defines a few functions to analyse and create ``stim.CircuitInstruction`` instances."""

import stim

COMBINED_RESET_AND_MEASUREMENT_INSTRUCTION_NAMES: frozenset[str] = frozenset(
    ["MR", "MRX", "MRY", "MRZ"]
)

SINGLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES: frozenset[str] = (
    frozenset(["M", "MX", "MY", "MZ"]) | COMBINED_RESET_AND_MEASUREMENT_INSTRUCTION_NAMES
)
MULTIPLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES: frozenset[str] = frozenset(
    ["MXX", "MYY", "MZZ", "MPP"]
)

MEASUREMENT_INSTRUCTION_NAMES = (
    SINGLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES | MULTIPLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES
)

ANNOTATION_INSTRUCTION_NAMES: frozenset[str] = frozenset(
    ["DETECTOR", "MPAD", "OBSERVABLE_INCLUDE", "QUBIT_COORDS", "SHIFT_COORDS", "TICK"]
)

RESET_INSTRUCTION_NAMES: frozenset[str] = frozenset(["R", "RX", "RY", "RZ"])


def is_combined_reset_and_measurement_instruction(
    instruction: stim.CircuitInstruction,
) -> bool:
    """Check if ``instruction`` is a combined reset/measurement instruction (e.g., ``MR``)."""
    return instruction.name in COMBINED_RESET_AND_MEASUREMENT_INSTRUCTION_NAMES  # pragma: no cover


def is_measurement_instruction(instruction: stim.CircuitInstruction) -> bool:
    """Check if ``instruction`` is a measurement instruction (e.g., ``M``)."""
    return instruction.name in MEASUREMENT_INSTRUCTION_NAMES  # pragma: no cover


def is_single_qubit_measurement_instruction(
    instruction: stim.CircuitInstruction,
) -> bool:
    """Check if ``instruction`` is a single-qubit measurement instruction (e.g., ``M``)."""
    return instruction.name in SINGLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES


def is_multi_qubit_measurement_instruction(
    instruction: stim.CircuitInstruction,
) -> bool:
    """Check if ``instruction`` is a multi-qubit measurement instruction (e.g. ``MXX``)."""
    return instruction.name in MULTIPLE_QUBIT_MEASUREMENT_INSTRUCTION_NAMES


def is_annotation_instruction(instruction: stim.CircuitInstruction) -> bool:
    """Check if the provided ``instruction`` is an annotation (e.g., ``DETECTOR``)."""
    return instruction.name in ANNOTATION_INSTRUCTION_NAMES


def is_reset_instruction(instruction: stim.CircuitInstruction) -> bool:
    """Check if the provided ``instruction`` is a reset instruction (e.g., ``R``)."""
    return instruction.name in RESET_INSTRUCTION_NAMES  # pragma: no cover



import functools
from collections.abc import Callable
from typing import Protocol

from typing_extensions import override

from tqec.compile.blocks.block import Block
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.plaquettes import PlaquetteLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.compile.blocks.layers.composed.repeated import RepeatedLayer
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.compile.specs.base import CubeBuilder, CubeSpec, PipeBuilder, PipeSpec
from tqec.compile.specs.enums import SpatialArms
from tqec.compile.specs.library.generators.fixed_boundary import FixedBoundaryConventionGenerator
from tqec.computation.cube import Port, YHalfCube, ZXCube
from tqec.plaquette.compilation.base import IdentityPlaquetteCompiler, PlaquetteCompiler
from tqec.plaquette.plaquette import Plaquettes
from tqec.plaquette.rpng.translators.base import RPNGTranslator
from tqec.plaquette.rpng.translators.default import DefaultRPNGTranslator
from tqec.templates.base import RectangularTemplate
from tqec.utils.enums import Basis, Orientation
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D
from tqec.utils.scale import LinearFunction


class _PlaquettesGenerator(Protocol):
    def __call__(
        self, reversed: bool, reset: Basis | None, measurement: Basis | None, /
    ) -> Plaquettes: ...


def _get_block(
    z_basis: Basis | None,
    has_spatial_junction_in_timeslice: bool,
    template: RectangularTemplate,
    plaquettes_generator: _PlaquettesGenerator,
    repetitions: LinearFunction,
) -> Block:
    """Get the block implemented with the provided ``template`` and ``plaquettes_generator``.

    This helper function handles all the complexity linked to generating a :class:`.Block` instance
    for the fixed boundary convention, especially when a spatial junction needs to be implemented
    with alternating plaquettes where it handles the alternation correctly even in REPEAT blocks.

    Raises:
        TQECError: if ``repetitions.slope`` is not even. This function requires an even
            repetition slope because 1) the implementation for odd slopes is more complex and
            requires the implementation of more classes and 2) all the repetitions we have at the
            moment have an even slope.

    """
    # Naming convention: {f,b}{init,memory,meas}
    # f: forward
    # b: backward
    # init: initialisation plaquette (with data-qubit resets)
    # memory: memory plaquette (no data-qubit reset/measurement)
    # meas: measurement plaquette (with data-qubit measurements)
    finit = plaquettes_generator(False, z_basis, None)
    fmemory = plaquettes_generator(False, None, None)
    fmeas = plaquettes_generator(False, None, z_basis)
    bmemory = plaquettes_generator(True, None, None)
    bmeas = plaquettes_generator(True, None, z_basis)

    if not has_spatial_junction_in_timeslice:
        return Block(
            [
                PlaquetteLayer(template, finit),
                RepeatedLayer(PlaquetteLayer(template, fmemory), repetitions),
                PlaquetteLayer(template, fmeas),
            ]
        )
    # else
    # Here, we need to implement the block by alternating forward/backward
    # schedules. To do so, we need to repeat the REPEAT block body only half the
    # initial number of repetitions. Basically, we need to compute
    # `repetitions // 2` and `repetitions % 2`. The problem is that a simple
    # `LinearFunction` instance is not enough to encode the result of these
    # operations. See for example `LinearFunction(3, 0) // 2` that should give
    # `1` for `k=1`, `3` for `k=2` and `4` for `k=3`. These three points are
    # not forming a straight line, so a `LinearFunction` instance cannot
    # represent them. Circumventing this issue could be done by adding more
    # classes (e.g., ModFunction), but it seems simpler for the moment to just
    # raise on unsupported inputs.
    if repetitions.slope % 2 == 1:
        raise NotImplementedError(
            "Cannot have an odd slope for the number of repetitions when a "
            "spatial junction is present."
        )
    halved_repetitions = LinearFunction(repetitions.slope // 2, repetitions.offset // 2)
    remainder = repetitions.offset % 2
    loop_replacement: list[BaseLayer | BaseComposedLayer] = [
        RepeatedLayer(
            SequencedLayers([PlaquetteLayer(template, bmemory), PlaquetteLayer(template, fmemory)]),
            halved_repetitions,
        )
    ]
    if remainder == 1:  # Note that remainder can only be 0 or 1.
        loop_replacement.append(PlaquetteLayer(template, bmemory))

    return Block(
        [
            PlaquetteLayer(template, finit),
            SequencedLayers(loop_replacement),
            PlaquetteLayer(template, fmeas if remainder == 1 else bmeas),
        ]
    )


class FixedBoundaryCubeBuilder(CubeBuilder):
    def __init__(
        self, compiler: PlaquetteCompiler, translator: RPNGTranslator = DefaultRPNGTranslator()
    ) -> None:
        """Implement the :class:`.CubeBuilder` interface for the fixed boundary convention.

        This class provides an implementation following the fixed-boundary convention.
        This convention consists in the fact that 2-body stabilizers on the boundary
        of a logical qubit are always at an even position(counting clockwise from the
        top left corner starting at 1).

        Args:
            compiler: compiler used to compile :class:`.Plaquette` instances returned from the
                provided ``translator``.
            translator: translator to obtain :class:`.Plaquette` instances from a RPNG description.

        """
        self._generator = FixedBoundaryConventionGenerator(translator, compiler)

    def _get_template_and_plaquettes_generator(
        self, spec: CubeSpec
    ) -> tuple[RectangularTemplate, _PlaquettesGenerator]:
        assert isinstance(spec.kind, ZXCube)
        x = spec.kind.x
        if not spec.is_spatial:
            orientation = Orientation.HORIZONTAL if x == Basis.Z else Orientation.VERTICAL

            def _memory_plaquettes_generator(
                is_reversed: bool, r: Basis | None, m: Basis | None
            ) -> Plaquettes:
                return self._generator.get_memory_qubit_plaquettes(is_reversed, orientation, r, m)

            return (
                self._generator.get_memory_qubit_raw_template(),
                _memory_plaquettes_generator,
            )

        # else
        def _spatial_plaquettes_generator(
            is_reversed: bool, r: Basis | None, m: Basis | None
        ) -> Plaquettes:
            return self._generator.get_spatial_cube_qubit_plaquettes(
                x, spec.spatial_arms, is_reversed, r, m
            )

        return (
            self._generator.get_spatial_cube_qubit_raw_template(),
            _spatial_plaquettes_generator,
        )

    @override
    def __call__(self, spec: CubeSpec, block_temporal_height: LinearFunction) -> Block:
        """Instantiate a :class:`.Block` instance implementing the provided ``spec``."""
        return self._call_impl(spec, block_temporal_height)

    @functools.cache
    def _call_impl(self, spec: CubeSpec, block_temporal_height: LinearFunction) -> Block:
        kind = spec.kind
        if isinstance(kind, Port):
            raise TQECError("Cannot build a block for a Port.")
        elif isinstance(kind, YHalfCube):
            raise NotImplementedError("Y cube is not implemented.")
        template, pgen = self._get_template_and_plaquettes_generator(spec)
        return _get_block(
            z_basis=kind.z,
            has_spatial_junction_in_timeslice=spec.has_spatial_up_or_down_pipe_in_timeslice,
            template=template,
            plaquettes_generator=pgen,
            repetitions=block_temporal_height,
        )


class FixedBoundaryPipeBuilder(PipeBuilder):
    def __init__(
        self, compiler: PlaquetteCompiler, translator: RPNGTranslator = DefaultRPNGTranslator()
    ) -> None:
        """Implement the :class:`.PipeBuilder` interface for the fixed boundary convention.

        This class provides an implementation following the fixed-boundary convention.
        This convention consists in the fact that 2-body stabilizers on the boundary
        of a logical qubit are always at an even position(counting clockwise from the
        top left corner starting at 1).

        Args:
            compiler: compiler used to compile :class:`.Plaquette` instances returned from the
                provided ``translator``.
            translator: translator to obtain :class:`.Plaquette` instances from a RPNG description.

        """
        self._generator = FixedBoundaryConventionGenerator(translator, compiler)

    @override
    def __call__(self, spec: PipeSpec, block_temporal_height: LinearFunction) -> Block:
        """Instantiate a :class:`.Block` instance implementing the provided ``spec``."""
        return self._call_impl(spec, block_temporal_height)

    @functools.cache
    def _call_impl(self, spec: PipeSpec, block_temporal_height: LinearFunction) -> Block:
        if spec.pipe_kind.is_temporal:
            return self.get_temporal_pipe_block(spec)
        return self.get_spatial_pipe_block(spec, block_temporal_height)

    #######################
    #    TEMPORAL PIPE    #
    #######################
    def get_temporal_pipe_block(self, spec: PipeSpec) -> Block:
        """Return the block to implement a temporal pipe based on the provided ``spec``.

        Args:
            spec: description of the pipe that should be implemented by this
                method. Should be a temporal pipe.

        Raises:
            AssertionError: if ``spec`` does not represent a temporal pipe.

        Returns:
            the block to implement a temporal pipe based on the provided
            ``spec``.

        """
        assert spec.pipe_kind.is_temporal
        hadamard_transition = spec.pipe_kind.has_hadamard
        z_orientation = (
            Orientation.HORIZONTAL if spec.pipe_kind.x == Basis.Z else Orientation.VERTICAL
        )
        memory_template = self._generator.get_memory_qubit_raw_template()
        memory_plaquettes = self._generator.get_memory_qubit_plaquettes(
            False,
            z_orientation.flip() if hadamard_transition else z_orientation,
            None,
            None,
        )
        memory_layer = PlaquetteLayer(memory_template, memory_plaquettes)

        if hadamard_transition:
            hadamard_template = self._generator.get_temporal_hadamard_raw_template()
            hadamard_plaquettes = self._generator.get_temporal_hadamard_plaquettes(
                False, z_orientation
            )
            hadamard_layer = PlaquetteLayer(hadamard_template, hadamard_plaquettes)
            return Block([hadamard_layer, memory_layer])
        # Else, it is a regular temporal junction
        return Block([memory_layer for _ in range(2)])

    ##############################
    #       SPATIAL PIPE         #
    ##############################
    @staticmethod
    def _get_spatial_cube_arms(spec: PipeSpec) -> SpatialArms:
        """Return the arm(s) corresponding to the provided ``spec``.

        Args:
            spec: pipe specification to get the arm(s) from.

        Raises:
            TQECError: if the provided ``spec`` is not a spatial pipe.

        Returns:
            the :class:`~tqec.compile.specs.enums.SpatialArms` instance
            corresponding to the provided ``spec``. The returned flag contains
            either one or two flags. If two flags are returned, they should be
            on the same line (e.g., it cannot be ``SpatialArms.RIGHT | SpatialArms.UP``
            but can be ``SpatialArms.RIGHT | SpatialArms.LEFT``).

            For example, if ``SpatialArms.LEFT`` is in the returned ``SpatialArms``
            flag, that means that the pipe represented by the provided ``spec``
            is the left arm of a spatial cube.

        """
        assert spec.pipe_kind.is_spatial
        # Check that we do have a spatial junction.
        assert any(spec.is_spatial for spec in spec.cube_specs)
        u, v = spec.cube_specs
        pipedir = spec.pipe_kind.direction
        arms = SpatialArms.NONE
        if u.is_spatial:
            arms |= SpatialArms.RIGHT if pipedir == Direction3D.X else SpatialArms.DOWN
        if v.is_spatial:
            arms |= SpatialArms.LEFT if pipedir == Direction3D.X else SpatialArms.UP
        return arms

    def _get_spatial_cube_pipe_block(
        self, spec: PipeSpec, block_temporal_height: LinearFunction
    ) -> Block:
        x, y, z = spec.pipe_kind.x, spec.pipe_kind.y, spec.pipe_kind.z
        assert x is not None or y is not None
        spatial_boundary_basis: Basis = x if x is not None else y  # type: ignore
        # Get the plaquette indices mappings
        arms = FixedBoundaryPipeBuilder._get_spatial_cube_arms(spec)
        pipe_template = self._generator.get_spatial_cube_arm_raw_template(arms)

        def plaquettes_generator(is_reversed: bool, r: Basis | None, m: Basis | None) -> Plaquettes:
            return self._generator.get_spatial_cube_arm_plaquettes(
                spatial_boundary_basis, arms, spec.cube_specs, is_reversed, r, m
            )

        return _get_block(
            z,
            spec.has_spatial_up_or_down_pipe_in_timeslice,
            pipe_template,
            plaquettes_generator,
            block_temporal_height,
        )

    def _get_spatial_regular_pipe_template(self, spec: PipeSpec) -> RectangularTemplate:
        """Return the template needed to implement the pipe representing the provided ``spec``."""
        assert spec.pipe_kind.is_spatial
        match spec.pipe_kind.direction, spec.pipe_kind.has_hadamard:
            case Direction3D.X, False:
                return self._generator.get_memory_vertical_boundary_raw_template()
            case Direction3D.X, True:
                return self._generator.get_spatial_vertical_hadamard_raw_template()
            case Direction3D.Y, False:
                return self._generator.get_memory_horizontal_boundary_raw_template()
            case Direction3D.Y, True:
                return self._generator.get_spatial_horizontal_hadamard_raw_template()
            case _:
                raise TQECError("Spatial pipes cannot have a direction equal to Direction3D.Z.")

    def _get_spatial_regular_non_hadamard_pipe_plaquettes_factory(
        self, spec: PipeSpec
    ) -> Callable[[bool, Basis | None, Basis | None], Plaquettes]:
        if spec.pipe_kind.direction == Direction3D.X:
            # Pipe between two cubes aligned on the X axis
            z_observable_orientation = (
                Orientation.HORIZONTAL if spec.pipe_kind.y == Basis.X else Orientation.VERTICAL
            )
            return lambda is_reversed, r, m: (
                self._generator.get_memory_vertical_boundary_plaquettes(
                    is_reversed, z_observable_orientation, r, m
                )
            )
        # Else, pipe between two cubes aligned on the Y axis
        z_observable_orientation = (
            Orientation.HORIZONTAL if spec.pipe_kind.x == Basis.Z else Orientation.VERTICAL
        )
        return lambda is_reversed, r, m: self._generator.get_memory_horizontal_boundary_plaquettes(
            is_reversed, z_observable_orientation, r, m
        )

    def _get_spatial_regular_pipe_plaquettes_factory(
        self, spec: PipeSpec
    ) -> Callable[[bool, Basis | None, Basis | None], Plaquettes]:
        assert spec.pipe_kind.is_spatial
        if not spec.pipe_kind.has_hadamard:
            return self._get_spatial_regular_non_hadamard_pipe_plaquettes_factory(spec)
        # Else, a Hadamard pipe.
        pipe_in_x_direction = spec.pipe_kind.direction == Direction3D.X
        top_left_basis = (
            spec.pipe_kind.get_basis_along(Direction3D.Y, at_head=True)
            if pipe_in_x_direction
            else spec.pipe_kind.get_basis_along(Direction3D.X, at_head=True)
        )
        assert top_left_basis is not None
        if pipe_in_x_direction:
            # Hadamard pipe between two cubes aligned on the X axis
            return lambda is_reversed, r, m: (
                self._generator.get_spatial_vertical_hadamard_plaquettes(
                    top_left_basis, is_reversed, r, m
                )
            )
        # Else, Hadamard pipe between two cubes aligned on the Y axis
        return lambda is_reversed, r, m: self._generator.get_spatial_horizontal_hadamard_plaquettes(
            top_left_basis, is_reversed, r, m
        )

    def _get_spatial_regular_pipe_block(
        self, spec: PipeSpec, block_temporal_height: LinearFunction
    ) -> Block:
        assert all(not spec.is_spatial for spec in spec.cube_specs)
        plaquettes_factory = self._get_spatial_regular_pipe_plaquettes_factory(spec)
        template = self._get_spatial_regular_pipe_template(spec)
        z = spec.pipe_kind.get_basis_along(Direction3D.Z, at_head=False)
        assert z is not None, "Spatial pipe should have a basis in the Z direction."
        return _get_block(
            z,
            spec.has_spatial_up_or_down_pipe_in_timeslice,
            template,
            plaquettes_factory,
            block_temporal_height,
        )

    def get_spatial_pipe_block(
        self, spec: PipeSpec, block_temporal_height: LinearFunction
    ) -> Block:
        """Return a :class:`.Block` instance implementing the provided ``spec``."""
        assert spec.pipe_kind.is_spatial
        cube_specs = spec.cube_specs
        if cube_specs[0].is_spatial or cube_specs[1].is_spatial:
            return self._get_spatial_cube_pipe_block(spec, block_temporal_height)
        return self._get_spatial_regular_pipe_block(spec, block_temporal_height)


FIXED_BOUNDARY_CUBE_BUILDER = FixedBoundaryCubeBuilder(IdentityPlaquetteCompiler)
FIXED_BOUNDARY_PIPE_BUILDER = FixedBoundaryPipeBuilder(IdentityPlaquetteCompiler)



"""Defines two classes to represent measurements in a quantum circuit.

This module defines :class:`Measurement` to represent a unique measurement in
a quantum circuit and :class:`RepeatedMeasurement` to represent a unique
measurement within a `REPEAT` instruction in a quantum circuit.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, cast

import stim
from typing_extensions import override

from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap
from tqec.utils.exceptions import TQECError
from tqec.utils.instructions import (
    is_multi_qubit_measurement_instruction,
    is_single_qubit_measurement_instruction,
)
from tqec.utils.position import Shift2D


class AbstractMeasurement(ABC):
    """Base class to represent a measurement."""

    @abstractmethod
    def offset_spatially_by(self, x: int, y: int) -> AbstractMeasurement:
        """Return a new instance offset by the provided spatial coordinates.

        Args:
            x: first spatial dimension offset.
            y: second spatial dimension offset.

        Returns:
            a new instance with the specified offset from ``self``.

        """

    @abstractmethod
    def offset_temporally_by(self, t: int) -> AbstractMeasurement:
        """Return a new instance offset by the provided temporal coordinates.

        Args:
            t: temporal offset.

        Returns:
            a new instance with the specified offset from ``self``.

        """

    @abstractmethod
    def __repr__(self) -> str:
        """Python magic method to represent an instance as a string."""

    @abstractmethod
    def map_qubit(self, qubit_map: Mapping[GridQubit, GridQubit]) -> AbstractMeasurement:
        """Return a copy of ``self`` with qubits mapped according to the provided``qubit_map``.

        The returned instance represents a measurement on the qubit obtained from ``self.qubit`` and
        the provided ``qubit_map``.

        Args:
            qubit_map: a correspondence map for qubits.

        Returns:
            a new measurement instance with the mapped qubit.

        """


@dataclass(frozen=True)
class Measurement(AbstractMeasurement):
    """A unique representation for each measurement in a quantum circuit.

    This class aims at being able to represent measurements in a quantum circuit
    in a unique and easily usable way.

    Note:
        This is not a global representation as the ``offset`` is always
        relative to the end of the quantum circuit considered.

    Attributes:
        qubit: qubit on which the represented measurement is performed.
        offset: negative offset representing the number of measurements
            performed on the provided qubit after the represented measurement.
            A value of ``-1`` means that the represented measurement is the
            last one applied on ``qubit``.

    Raises:
        TQECError: if the provided ``offset`` is not strictly negative.

    """

    qubit: GridQubit
    offset: int

    def __post_init__(self) -> None:
        if self.offset >= 0:
            raise TQECError("Measurement.offset should be negative.")

    @override
    def offset_spatially_by(self, x: int, y: int) -> Measurement:
        return Measurement(self.qubit + Shift2D(x, y), self.offset)

    @override
    def offset_temporally_by(self, t: int) -> Measurement:
        return Measurement(self.qubit, self.offset + t)

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.qubit}, {self.offset})"

    def __str__(self) -> str:
        return f"M[{self.qubit},{self.offset}]"

    @override
    def map_qubit(self, qubit_map: Mapping[GridQubit, GridQubit]) -> Measurement:
        return Measurement(qubit_map[self.qubit], self.offset)

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the measurement.

        Returns:
            a dictionary with the keys ``qubit`` and ``offset`` and their
            corresponding values.

        """
        return {"qubit": [self.qubit.x, self.qubit.y], "offset": self.offset}

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Measurement:
        """Return a measurement from its dictionary representation.

        Args:
            data: dictionary with the keys ``qubit`` and ``offset``.

        Returns:
            a new instance of :class:`Measurement` with the provided
            ``qubit`` and ``offset``.

        """
        qubit = GridQubit(data["qubit"][0], data["qubit"][1])
        offset = data["offset"]
        return Measurement(qubit, offset)


def get_measurements_from_circuit(circuit: stim.Circuit) -> list[Measurement]:
    """Get all the measurements found in the provided circuit.

    Args:
        circuit: circuit to extract measurements from.

    Raises:
        TQECError: if the provided circuit contains a ``REPEAT`` block.
        TQECError: if the provided circuit contains a multi-qubit measurement
            gate such as ``MXX`` or ``MPP``.
        TQECError: if the provided circuit contains a single-qubit
            measurement gate with a non-qubit target.

    Returns:
        all the measurements present in the provided ``circuit``, in their order
        of appearance (so in increasing order of measurement record offsets).

    """
    qubit_map = QubitMap.from_circuit(circuit)
    num_measurements: dict[GridQubit, int] = {}
    measurements_reverse_order: list[Measurement] = []
    for instruction in reversed(circuit):
        if isinstance(instruction, stim.CircuitRepeatBlock):
            raise TQECError(
                "Found a REPEAT block in get_measurements_from_circuit. This is not supported."
            )
        if is_multi_qubit_measurement_instruction(instruction):
            raise TQECError(
                f"Got a multi-qubit measurement instruction ({instruction.name}) "
                "but multi-qubit measurements are not supported yet."
            )
        if is_single_qubit_measurement_instruction(instruction):
            for (target,) in reversed(instruction.target_groups()):
                if not target.is_qubit_target:
                    raise TQECError(
                        "Found a measurement instruction with a target that is "
                        f"not a qubit target: {instruction}."
                    )
                qi: int = cast(int, target.qubit_value)
                qubit = qubit_map.i2q[qi]
                meas_index_on_qubit = num_measurements.get(qubit, 0) + 1
                num_measurements[qubit] = meas_index_on_qubit
                measurements_reverse_order.append(Measurement(qubit, -meas_index_on_qubit))
    return measurements_reverse_order[::-1]



from abc import ABC, abstractmethod

from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.utils.exceptions import TQECError


class CompilationPass(ABC):
    """Base interface that should be implemented by all compilation passes."""

    @abstractmethod
    def run(self, circuit: ScheduledCircuit, check_all_flows: bool = False) -> ScheduledCircuit:
        """Run the compilation pass.

        Args:
            circuit: quantum circuit that should be modified by the compilation
                pass.
            check_all_flows: if ``True``, this method will check if the final
                compiled circuit has the same Pauli flows as the original
                circuit. This check may be costly in terms of runtime, so it is
                disabled by default.

        Returns:
            the compiled quantum circuit.

        """
        pass

    def check_flows(
        self, original_circuit: ScheduledCircuit, modified_circuit: ScheduledCircuit
    ) -> None:
        """Check that the two provided circuits have the exact same flows.

        This method can be used to check if two quantum circuits are functionally
        equivalent. It lists all the flows of the provided ``original_circuit``
        and checks that ``modified_circuit`` contains all these flows.

        Args:
            original_circuit: first circuit, supposed to be the circuit before
                applying the compilation pass.
            modified_circuit: second circuit, supposed to be the circuit after
                applying the compilation pass.

        Raises:
            TQECError: if the two provided circuits are not functionally
                equivalent (i.e. ``modified_circuit`` does not have at least
                one of flow of ``original_circuit``).

        """
        original_flows = original_circuit.get_circuit().flow_generators()
        if not modified_circuit.get_circuit().has_all_flows(original_flows):
            raise TQECError("Modified circuit does not contain")



import os
from pathlib import Path
from typing import Final

from platformdirs import user_data_path

USER_DATA_PATH: Final[Path] = user_data_path(appname="TQEC")


# Get the detector database path.
# This path can be provided through the TQEC_DETECTOR_DATABASE_PATH environment variable, or it
# defaults to a user-controlled directory.
def _get_database_path() -> Path:
    if (env_db_path := os.getenv("TQEC_DETECTOR_DATABASE_PATH")) is not None:
        return Path(env_db_path)  # pragma: no cover
    else:
        return USER_DATA_PATH / "detector_database.pkl"


DEFAULT_DETECTOR_DATABASE_PATH: Final[Path] = _get_database_path()

PKG_DIR: Final[Path] = Path(__file__).parent.parent
GALLERY_DAE_DIR: Final[Path] = PKG_DIR / "gallery" / "dae"

# Create the directories that are used in case they do not exist.
_DIRECTORIES_TO_CREATE: Final[list[Path]] = [USER_DATA_PATH]
for directory in _DIRECTORIES_TO_CREATE:
    if not directory.exists():
        directory.mkdir(parents=True)



from tqec.circuit.measurement_map import MeasurementRecordsMap
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.observables.abstract_observable import AbstractObservable
from tqec.compile.observables.builder import (
    ObservableBuilder,
    ObservableComponent,
    get_observable_with_measurement_records,
)
from tqec.compile.tree.node import LayerNode


def _get_ordered_leaves(root: LayerNode) -> list[LayerNode]:
    """Return the leaves of the tree in time order."""
    if root.is_leaf:
        return [root]
    return [n for child in root.children for n in _get_ordered_leaves(child)]


def annotate_observable(
    root: LayerNode,
    k: int,
    observable: AbstractObservable,
    observable_index: int,
    observable_builder: ObservableBuilder,
) -> None:
    """Annotates the observables on the tree.

    Args:
        root: root node of the tree.
        k: distance parameter.
        observable: observable to annotate.
        observable_index: index of the observable in the circuit.
        observable_builder: builder that computes and constructs qubits whose
            measurements will be included in the logical observable.

    """
    for z, subtree_root in enumerate(root.children):
        leaves = _get_ordered_leaves(subtree_root)
        obs_slice = observable.slice_at_z(z)
        # Annotate the observable at the bottom of the blocks
        _annotate_observable_at_node(
            leaves[0],
            obs_slice,
            k,
            observable_index,
            observable_builder,
            ObservableComponent.BOTTOM_STABILIZERS,
        )
        readout_layer = leaves[-1]
        if obs_slice.temporal_hadamard_pipes:
            readout_layer = leaves[-2]
            # Annotate the observable at the realignment layer in temporal hadamard pipes
            _annotate_observable_at_node(
                leaves[-1],
                obs_slice,
                k,
                observable_index,
                observable_builder,
                ObservableComponent.REALIGNMENT,
            )
        # Annotate the observable at the top of the blocks
        _annotate_observable_at_node(
            readout_layer,
            obs_slice,
            k,
            observable_index,
            observable_builder,
            ObservableComponent.TOP_READOUTS,
        )


def _annotate_observable_at_node(
    node: LayerNode,
    obs_slice: AbstractObservable,
    k: int,
    observable_index: int,
    observable_builder: ObservableBuilder,
    component: ObservableComponent,
) -> None:
    circuit = node.get_annotations(k).circuit
    assert circuit is not None
    measurement_record = MeasurementRecordsMap.from_scheduled_circuit(circuit)
    assert isinstance(node._layer, LayoutLayer)
    template, _ = node._layer.to_template_and_plaquettes()
    obs_qubits = observable_builder.build(k, template, obs_slice, component)
    if obs_qubits:
        obs_annotation = get_observable_with_measurement_records(
            obs_qubits, measurement_record, observable_index
        )
        node.get_annotations(k).observables.append(obs_annotation)



"""Implements a wrapper to standardise the stim coordinate system across the code base."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class StimCoordinates:
    """Wrapper around a tuple of coordinates.

    This class should be used whenever a ``stim.CircuitInstruction`` that expects
    arguments that are interpreted by stim as coordinates is built. This is the
    case of the ``DETECTOR`` instruction for example.

    """

    _ABS_TOL = 1e-10

    x: float
    """First spatial coordinate."""
    y: float
    """Second spatial coordinate."""
    t: float | None = None
    """Third coordinate, most of the time associated to time, but might be associated to the third
    spatial coordinate in some contexts.
    """

    def to_stim_coordinates(self) -> tuple[float, ...]:
        """Return a tuple that can be used as ``stim`` coordinates.

        The output of this method is intended to be forwarded to the third
        argument of ``stim.CircuitInstruction``.

        """
        if self.t is not None:
            return (self.x, self.y, self.t)
        return (self.x, self.y)

    def __str__(self) -> str:  # pragma: no cover
        return "(" + ",".join(f"{c:.2f}" for c in self.to_stim_coordinates()) + ")"

    def offset_spatially_by(self, x: float, y: float) -> StimCoordinates:
        """Return a new instance with offset spatial coordinates."""
        return StimCoordinates(self.x + x, self.y + y, self.t)

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, StimCoordinates)
            and abs(self.x - value.x) < StimCoordinates._ABS_TOL
            and abs(self.y - value.y) < StimCoordinates._ABS_TOL
            and (
                (self.t is None and value.t is None)
                or (
                    self.t is not None
                    and value.t is not None
                    and abs(self.t - value.t) < StimCoordinates._ABS_TOL
                )
            )
        )

    def __hash__(self) -> int:
        return hash((self.x, self.y, self.t))  # pragma: no cover

    def __lt__(self, other: StimCoordinates) -> bool:
        return self.to_stim_coordinates() < other.to_stim_coordinates()

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the coordinates.

        The dictionary is intended to be used as a JSON object.

        """
        return {
            "x": self.x,
            "y": self.y,
            "t": self.t,
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> StimCoordinates:
        """Return a new instance of :class:`StimCoordinates` from its dictionary representation.

        Args:
            data: dictionary with the keys ``x``, ``y`` and ``t``.

        Returns:
            a new instance of :class:`StimCoordinates` with the provided
            ``x``, ``y`` and ``t``.

        """
        x = data["x"]
        y = data["y"]
        t = data.get("t")
        return StimCoordinates(x, y, t)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Final

from typing_extensions import Self

from tqec.compile.blocks.enums import SpatialBlockBorder
from tqec.utils.position import PhysicalQubitShape2D
from tqec.utils.scale import PhysicalQubitScalable2D

EXPECTED_SPATIAL_BORDER_WIDTH: Final[int] = 2
"""Hard-coded expected spatial border width in qubit coordinates.

At the moment, we need to ensure that removing one spatial border from any given layer will trim out
a band of qubits with a known width. Some computations in the code base indirectly depends on the
fact that this value is 2 for historical reasons.

Even though we do not need to in the foreseeable future, changing that width will likely lead to
various errors in the code base.

"""


class WithSpatialFootprint(ABC):
    """Base class providing the interface implemented by objects that have a spatial footprint."""

    def __init__(self, trimmed_spatial_borders: frozenset[SpatialBlockBorder] = frozenset()):
        """Initialise the instance.

        Args:
            trimmed_spatial_borders: all the spatial borders that have been trimmed from the layer.

        """
        super().__init__()
        self._trimmed_spatial_borders = trimmed_spatial_borders

    @property
    @abstractmethod
    def scalable_shape(self) -> PhysicalQubitScalable2D:
        """Return the 2-dimensional shape of the object.

        Note:
            This method should return the shape in qubit-coordinates. That means
            that usual plaquette for the rotated surface code measuring 4-body
            stabilizers (i.e., the square ones with a data-qubit on each corner
            and a syndrome qubit in the middle of the square) has a shape of
            ``(3, 3)``.

        Returns:
            the 2-dimensional shape **in qubit-coordinates** of the object as an
            exact expression that can then be used to compute the shape for any
            value of ``k``.

        """
        pass

    def shape(self, k: int) -> PhysicalQubitShape2D:
        """Return the 2-dimensional shape of the object for the given ``k``.

        Args:
            k: scaling parameter.

        Returns:
            the 2-dimensional shape of the object for the given ``k``.

        """
        return self.scalable_shape.to_shape_2d(k)  # pragma: no cover

    @abstractmethod
    def with_spatial_borders_trimmed(self, borders: Iterable[SpatialBlockBorder]) -> Self:
        """Return ``self`` with the provided spatial borders removed.

        Args:
            borders: spatial borders to remove.

        Returns:
            a copy of ``self`` with the provided ``borders`` removed.

        """
        pass

    @property
    def trimmed_spatial_borders(self) -> frozenset[SpatialBlockBorder]:
        """Get the spatial layers that have been trimmed off from ``self``."""
        return self._trimmed_spatial_borders



"""Describe the location of the observable measurements in the block graph."""

from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from typing import TYPE_CHECKING

from tqec.compile.specs.enums import SpatialArms
from tqec.computation.block_graph import BlockGraph
from tqec.computation.correlation import CorrelationSurface, ZXEdge
from tqec.computation.cube import Cube, ZXCube
from tqec.computation.pipe import Pipe
from tqec.utils.enums import Basis, Pauli
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D, Position3D

if TYPE_CHECKING:
    from pyzx.graph.graph_s import GraphS


@dataclass(frozen=True)
class CubeWithArms:
    """A cube with its arms in the block graph.

    The arms are used to specify the connectivity of a spatial cube in the graph
    or in a correlation surface. A regular cube should always have `arms` set to
    `SpatialArms.NONE`.

    Attributes:
        cube: The cube in the block graph.
        arms: The arms of the cube in the block graph or correlation surface.
            If the cube is not spatial, the arms should be set to
            `SpatialArms.NONE`.

    """

    cube: Cube
    arms: SpatialArms = SpatialArms.NONE

    def __post_init__(self) -> None:
        if self.arms != SpatialArms.NONE and not self.cube.is_spatial:
            raise TQECError(
                "The `arms` attribute should be `SpatialArms.NONE` for non-spatial cubes."
            )


@dataclass(frozen=True)
class PipeWithArms:
    """A pipe with the arms of the two cubes it connects in the block graph.

    Attributes:
        pipe: The pipe in the block graph.
        cube_arms: A tuple of arms of the two cubes (``u`` and ``v``) the pipe
            connects. If the cubes are not spatial, the arms should be set to
            `SpatialArms.NONE`.

    """

    pipe: Pipe
    cube_arms: tuple[SpatialArms, SpatialArms] = (SpatialArms.NONE, SpatialArms.NONE)


@dataclass(frozen=True)
class PipeWithObservableBasis:
    """A temporal Hadamard pipe with the attached observable basis at its head.

    Attributes:
        pipe: The temporal Hadamard pipe in the block graph.
        observable_basis: The basis of the correlation surface at the head of
            the pipe.

    """

    pipe: Pipe
    observable_basis: Basis

    def __post_init__(self) -> None:
        if not self.pipe.kind.has_hadamard or self.pipe.direction != Direction3D.Z:
            raise TQECError("The ``pipe`` should be a temporal Hadamard pipe.")


@dataclass(frozen=True)
class AbstractObservable:
    """An abstract description of a logical observable in the
    :py:class:`~tqec.computation.BlockGraph`.

    A logical observable corresponds to all the
    `OBSERVABLE_INCLUDE <https://github.com/quantumlib/Stim/blob/main/doc/gates.md#OBSERVABLE_INCLUDE>`_
    instructions with the same observable index in the ``stim`` circuit
    and is composed of a set of measurements. Abstract observable
    specifies where are the measurements located in the block graph.

    Attributes:
        top_readout_cubes: A set of cubes of which a line of data qubit readouts
            on the top face should be included in the observable.
        top_readout_pipes: A set of pipes of which a single data qubit readout
            on the top face might be included in the observable. The data qubit
            is on the center of the interface between the two cubes. The other
            data qubit readouts that may be included in the observable will be
            handled by the ``top_readout_cubes`` set.
        bottom_stabilizer_cubes: A set of spatial cubes of which the stabilizer
            measurements on the bottom face should be included in the observable.
            Usually, this is only used for the single-cube stability experiment
            where no pipes are involved. In other cases, the stabilizer measured
            at the bottom face of the cubes will be handled by the pipes
            connecting them, i.e., the ``bottom_stabilizer_pipes`` set.
        bottom_stabilizer_pipes: A set of pipes of which a region of stabilizer
            measurements on the bottom face, which actually takes place in the
            cubes it connects, should be included in the observable.
        temporal_hadamard_pipes: A set of temporal Hadamard pipes of which a
            single stabilizer measurements at the realignment layer represented
            by the pipe might be included in the logical observable. It is only
            relevant for the fixed-bulk convention, where the temporal Hadamard
            includes a realignment layer of stabilizers.

    """

    top_readout_cubes: frozenset[CubeWithArms] = frozenset()
    top_readout_pipes: frozenset[PipeWithArms] = frozenset()
    bottom_stabilizer_cubes: frozenset[CubeWithArms] = frozenset()
    bottom_stabilizer_pipes: frozenset[PipeWithArms] = frozenset()
    temporal_hadamard_pipes: frozenset[PipeWithObservableBasis] = frozenset()

    def slice_at_z(self, z: int) -> AbstractObservable:
        """Get the observable slice at the given z position."""
        return AbstractObservable(
            frozenset(c for c in self.top_readout_cubes if c.cube.position.z == z),
            frozenset(p for p in self.top_readout_pipes if p.pipe.u.position.z == z),
            frozenset(c for c in self.bottom_stabilizer_cubes if c.cube.position.z == z),
            frozenset(p for p in self.bottom_stabilizer_pipes if p.pipe.u.position.z == z),
            frozenset(p for p in self.temporal_hadamard_pipes if p.pipe.u.position.z == z),
        )


def compile_correlation_surface_to_abstract_observable(
    block_graph: BlockGraph,
    correlation_surface: CorrelationSurface,
    include_temporal_hadamard_pipes: bool = False,
) -> AbstractObservable:
    """Compile a ``CorrelationSurface`` into an ``AbstractObservable`` in the block graph.

    The correlation surface translates into measurements to be included in the
    observable in the following ways:

    1. The surface attaches to the top face of some blocks. This means that part
    of the logical operator is measured by reading the data qubits. The parity
    change must be accounted for, and the measurement results should be included
    in the tracked logical observable.

    2. The surface spans the XY plane within some blocks. This represents a
    region of stabilizer measurements in the basis of the surface, whose products
    give the parity of the logical operators on the surface edges. Stabilizer
    measurements need to be included in the tracked logical observable to account
    for the correlation between logical operators at different spatial locations.
    We choose the stabilizer measurements at the first layer (i.e., the earliest
    in time or the bottom face of the block), since with sufficiently advanced
    software we will have greater confidence in these measurements earlier and
    in principle be able to make decisions based on these measurements earlier.

    3. A temporal Hadamard pipe under fixed-bulk convention includes a layer of
    realignment stabilizers, which might includes a single stabilizer measurements
    that need to be added to the observable.

    The compilation process is as follows:

    1. Find all the spatial cubes involved in the correlation surface. For
    each cube:

    - If a surface is perpendicular to the XY plane, include data qubit readouts
    on the top face of the cube in the observable. Correlation surfaces
    parallel to the cube's normal direction are guaranteed to attach to an
    even number of arms.
        - If exactly two arms touch the surface, add the cube and arms to the
        ``top_readout_cubes`` set.
        - If four arms touch the surface, split the arms into two pairs (e.g.
        ``SpatialArms.LEFT | SpatialArms.DOWN`` and
        ``SpatialArms.RIGHT | SpatialArms.UP``), and add the cube and arms
        to the ``top_readout_cubes`` set.

    2. Iterate over all the edges in the correlation surface. For each edge:
    - If the edge is vertical, check if the surface is attached to the top face
    of the top cube. If so, add the top cube to the ``top_readout_cubes`` set.
    If the edge is a hadamard edge, add the pipe to the ``temporal_hadamard_pipes``
    set.
    - If the edge is horizontal, check if the surface is attached to the top face
    of the pipe. If so, add the pipe to the ``top_readout_pipes`` set; otherwise,
    add the pipe to the ``bottom_stabilizer_pipes`` set.
    - For each cube in the pipe, check if the surface is attached to the top face
    of the cube. If so, add the cube to the ``top_readout_cubes`` set. Otherwise,
    add the pipe to the ``bottom_stabilizer_pipes`` set with the arms of the
    cubes in the pipe.

    Args:
        block_graph: The block graph whose corresponding ZX graph supports the
            correlation surface.
        correlation_surface: The correlation surface to convert into an abstract
            observable.
        include_temporal_hadamard_pipes: whether to include the temporal hadamard
            pipes in the observable. This is only relevant for the fixed bulk
            convention.

    Returns:
        The abstract observable corresponding to the correlation surface in the block graph.

    Raises:
        TQECError: If the block graph has open ports or the block graph cannot
            support the correlation surface.

    """
    # 0. Handle single node edge case
    if correlation_surface.is_single_node:
        # single stability experiment
        cube = block_graph.cubes[0]
        cube_with_arms = CubeWithArms(cube)
        if cube.is_spatial:
            return AbstractObservable(bottom_stabilizer_cubes=frozenset([cube_with_arms]))
        # single memory experiment
        return AbstractObservable(top_readout_cubes=frozenset([cube_with_arms]))

    pg = block_graph.to_zx_graph()
    _check_correlation_surface_validity(correlation_surface, pg.g)

    endpoints_to_edge: dict[frozenset[Position3D], list[ZXEdge]] = {}
    for edge in correlation_surface.span:
        u, v = edge.u.id, edge.v.id
        endpoints = frozenset({pg[u], pg[v]})
        endpoints_to_edge.setdefault(endpoints, []).append(edge)

    top_readout_cubes: set[CubeWithArms] = set()
    top_readout_pipes: set[PipeWithArms] = set()
    bottom_stabilizer_pipes: set[PipeWithArms] = set()
    temporal_hadamard_pipes: set[PipeWithObservableBasis] = set()

    # 1. Handle spatial cubes top readouts
    for node in correlation_surface.span_vertices:
        cube = block_graph[pg[node]]
        if not cube.is_spatial:
            continue

        kind = cube.kind
        assert isinstance(kind, ZXCube)
        bases = correlation_surface.bases_at(node)
        normal_basis = kind.normal_basis
        # correlation surface parallel to the normal direction of the cube
        # accounts for the top readout measurements
        # we need to record the arm flags to specify different shapes of the
        # observable lines, e.g. L-shape, 7-shape, -- shape, etc.
        if {normal_basis.flipped()} != bases:
            # check correlation edges in the cube arms
            arms = SpatialArms.NONE
            for arm, shift in SpatialArms.get_map_from_arm_to_shift().items():
                edges = endpoints_to_edge.get(
                    frozenset({cube.position, cube.position.shift_by(*shift)})
                )
                if edges is not None and any(
                    n.basis == normal_basis for edge in edges for n in edge
                ):
                    arms |= arm
            assert len(arms) in {
                2,
                4,
            }, "The correlation parity should be even for parallel correlation surface."
            # Two separate lines in the cube
            # By convention, we always split the four arms into [LEFT | DOWN] and [RIGHT | UP]
            if len(arms) == 4:
                top_readout_cubes.add(CubeWithArms(cube, SpatialArms.LEFT | SpatialArms.DOWN))
                top_readout_cubes.add(CubeWithArms(cube, SpatialArms.RIGHT | SpatialArms.UP))
            else:
                top_readout_cubes.add(CubeWithArms(cube, arms))

    # 2. Handle all the pipes
    def has_obs_include(cube: Cube, correlation: Basis) -> bool:
        """Check if the top data qubit readout should be included in the observable."""
        if cube.is_y_cube:
            return True
        assert isinstance(cube.kind, ZXCube)
        # No pipe at the top
        if block_graph.has_pipe_between(cube.position, cube.position.shift_by(0, 0, 1)):
            return False
        # The correlation surface must be attached to the top face
        return cube.kind.z.value == correlation.value

    for edge in correlation_surface.span:
        up, vp = pg[edge.u.id], pg[edge.v.id]
        pipe = block_graph.get_pipe(up, vp)
        # Vertical pipes
        if pipe.direction == Direction3D.Z:
            # Temporal Hadamard might have measurements that should be included
            # during realignment of plaquettes under fixed-bulk convention
            if include_temporal_hadamard_pipes and pipe.kind.has_hadamard:
                temporal_hadamard_pipes.add(PipeWithObservableBasis(pipe, edge.u.basis))
            if has_obs_include(pipe.v, edge.v.basis):
                top_readout_cubes.add(CubeWithArms(pipe.v))
            continue
        arms_u = (
            SpatialArms.from_cube_in_graph(pipe.u, block_graph)
            if pipe.u.is_spatial
            else SpatialArms.NONE
        )
        arms_v = (
            SpatialArms.from_cube_in_graph(pipe.v, block_graph)
            if pipe.v.is_spatial
            else SpatialArms.NONE
        )
        # Horizontal pipes
        pipe_top_face = pipe.kind.z
        assert pipe_top_face is not None, "The pipe is guaranteed to be spatial."
        # There is correlation surface attached to the top of the pipe
        if pipe_top_face.value == edge.u.basis.value:
            top_readout_pipes.add(PipeWithArms(pipe, (arms_u, arms_v)))
            for cube, n in zip(pipe, edge):
                # Spatial cubes have already been handled
                if cube.is_spatial:
                    continue
                if has_obs_include(cube, n.basis):
                    top_readout_cubes.add(CubeWithArms(cube))
        else:
            bottom_stabilizer_pipes.add(PipeWithArms(pipe, (arms_u, arms_v)))

    return AbstractObservable(
        top_readout_cubes=frozenset(top_readout_cubes),
        top_readout_pipes=frozenset(top_readout_pipes),
        bottom_stabilizer_pipes=frozenset(bottom_stabilizer_pipes),
        temporal_hadamard_pipes=frozenset(temporal_hadamard_pipes),
    )


def _check_correlation_surface_validity(correlation_surface: CorrelationSurface, g: GraphS) -> None:
    # Needs to be imported here to avoid pulling pyzx when importing this module.
    from tqec.interop.pyzx.utils import zx_to_pauli  # noqa: PLC0415

    """Check the ZX graph can support the correlation surface."""
    # 1. Check the vertices in the correlation surface are in the graph
    if missing_vertices := (correlation_surface.span_vertices - g.vertex_set()):
        raise TQECError(
            "The following vertices in the correlation surface are "
            f"not in the graph: {missing_vertices} "
        )
    # 2. Check the edges in the correlation surface are in the graph
    edges = g.edge_set()  # type: ignore
    for edge in correlation_surface.span:
        e = (edge.u.id, edge.v.id)
        if e not in edges and (e[1], e[0]) not in edges:
            raise TQECError(f"Edge {e} in the correlation surface is not in the graph.")
    # 3. Check parity around each vertex
    for v in correlation_surface.span_vertices:
        pauli = zx_to_pauli(g, v)
        edges = correlation_surface.edges_at(v)
        match pauli:
            case Pauli.I:
                continue
            case Pauli.Y:
                # Y vertex should have Y pauli
                if len(edges) != 2 or len(correlation_surface.bases_at(v)) != 2:
                    raise TQECError(
                        f"Y type vertex should have Pauli Y supported on it, vertex {v} violates"
                        " the rule."
                    )
                continue
            case _:
                counts = Counter(edge.u.basis if edge.u.id == v else edge.v.basis for edge in edges)
                v_basis = pauli.to_basis()
                if counts[v_basis.flipped()] not in [0, len(g.incident_edges(v))]:
                    raise TQECError(
                        "X (Z) type vertex should have Pauli Z (X) Pauli supported on "
                        f" all or no edges, vertex {v} violates the rule."
                    )
                if counts[v_basis] % 2 != 0:
                    raise TQECError(
                        f"X (Z) type vertex should have even number of Pauli X (Z) supported"
                        f" on the edges, vertex {v} violates the rule."
                    )



from enum import Enum

from tqec.templates.enums import TemplateBorder
from tqec.utils.position import Direction3D, SignedDirection3D


class SpatialBlockBorder(Enum):
    """Enumerates the 4 different spatial borders for a block."""

    X_NEGATIVE = SignedDirection3D(Direction3D.X, False)
    X_POSITIVE = SignedDirection3D(Direction3D.X, True)
    Y_NEGATIVE = SignedDirection3D(Direction3D.Y, False)
    Y_POSITIVE = SignedDirection3D(Direction3D.Y, True)

    def to_template_border(self) -> TemplateBorder:
        """Return the template border corresponding to ``self``."""
        match self:
            case SpatialBlockBorder.X_NEGATIVE:
                return TemplateBorder.LEFT
            case SpatialBlockBorder.X_POSITIVE:
                return TemplateBorder.RIGHT
            case SpatialBlockBorder.Y_NEGATIVE:
                return TemplateBorder.TOP
            case SpatialBlockBorder.Y_POSITIVE:
                return TemplateBorder.BOTTOM
            # add a wildcard pattern when function returns `None`
            # flagged by ty
            case _:
                raise ValueError(f"Cannot return the template border corresponding to {self} .")


class TemporalBlockBorder(Enum):
    """Enumerates the 2 different temporal borders for a block."""

    Z_NEGATIVE = SignedDirection3D(Direction3D.Z, False)
    Z_POSITIVE = SignedDirection3D(Direction3D.Z, True)


def border_from_signed_direction(
    direction: SignedDirection3D,
) -> SpatialBlockBorder | TemporalBlockBorder:
    """Get the block border from its direction.

    Args:
        direction: direction indicating the border to return.

    Returns:
        The border corresponding to the provided direction. Imagining a line
        starting from the block center and extending in the provided direction,
        the border pierced by the line is returned by this function.

        For example, if ``"+X"`` is provided, the returned border will be
        ``SpatialBlockBorder.X_POSITIVE``

    """
    match direction:
        case SignedDirection3D(Direction3D.Z, _):
            return TemporalBlockBorder(direction)
        case _:
            return SpatialBlockBorder(direction)



from __future__ import annotations

from collections.abc import Iterable
from functools import cached_property
from typing import Final, TypeGuard

from typing_extensions import override

from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.compile.blocks.enums import SpatialBlockBorder
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.plaquettes import PlaquetteLayer
from tqec.compile.blocks.positioning import (
    LayoutCubePosition2D,
    LayoutPipePosition2D,
    LayoutPosition2D,
)
from tqec.compile.generation import generate_circuit
from tqec.plaquette.plaquette import Plaquettes
from tqec.templates.enums import TemplateBorder
from tqec.templates.layout import LayoutTemplate
from tqec.utils.exceptions import TQECError
from tqec.utils.position import BlockPosition2D, Direction3D, Shift2D
from tqec.utils.scale import LinearFunction, PhysicalQubitScalable2D

DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER: Final[int] = 1
"""Default number of qubits that are shared between two neighbouring layers."""


def contains_only_plaquette_layers(
    layers: dict[LayoutPosition2D, BaseLayer],
) -> TypeGuard[dict[LayoutPosition2D, PlaquetteLayer]]:
    """Ensure correct typing when used in a conditional block."""
    return all(isinstance(layer, PlaquetteLayer) for layer in layers.values())


class LayoutLayer(BaseLayer):
    def __init__(
        self,
        layers: dict[LayoutPosition2D, BaseLayer],
        element_shape: PhysicalQubitScalable2D,
    ) -> None:
        """Glue several other layers together on a 2-dimensional grid.

        Args:
            layers: a mapping from positions on the 2-dimensional space to
                blocks implementing the circuit that should be present at that
                position.
                The mapping is expected to represent a connected computation.
            element_shape: scalable shape (in qubit coordinates) of each entry
                in the provided ``layers``.

        Raises:
            TQECError: if ``layers`` is empty.
            TQECError: if ``trimmed_spatial_borders`` is not empty.

        """
        super().__init__(frozenset())
        self._layers = layers
        self._element_shape = element_shape
        self._post_init_check()

    def _post_init_check(self) -> None:
        if not self.layers:
            raise TQECError(f"An instance of {type(self).__name__} should have at least one layer.")
        if self.trimmed_spatial_borders:
            raise TQECError(f"{LayoutLayer.__name__} cannot have trimmed spatial borders.")

    @property
    def layers(self) -> dict[LayoutPosition2D, BaseLayer]:
        """Return the layers composing ``self``."""
        return self._layers

    @property
    def element_shape(self) -> PhysicalQubitScalable2D:
        """Return the scalable shape of each stored elements."""
        return self._element_shape

    @cached_property
    def bounds(self) -> tuple[BlockPosition2D, BlockPosition2D]:
        """Get the top-left and bottom-right corners of the bounding box of ``self``.

        Returns:
            a tuple containing the corners of ``self``'s bounding box as positions
            containing the two minimum (top-left corner) or maximum (bottom-right
            corner) coordinates found in ``self``.

        """
        xs = [pos._x for pos in self.layers.keys()]
        ys = [pos._y for pos in self.layers.keys()]
        minx, maxx = min(xs), max(xs)
        miny, maxy = min(ys), max(ys)
        # We know for sure that the minima and maxima above are located on cube
        # positions and so are multiples of 2.
        return (
            BlockPosition2D(minx // 2, miny // 2),
            BlockPosition2D(maxx // 2, maxy // 2),
        )

    @property
    @override
    def scalable_shape(self) -> PhysicalQubitScalable2D:
        minp, maxp = self.bounds
        shapex, shapey = (maxp.x - minp.x) + 1, (maxp.y - minp.y) + 1
        return PhysicalQubitScalable2D(
            shapex * (self.element_shape.x - DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER)
            + DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER,
            shapey * (self.element_shape.y - DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER)
            + DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER,
        )

    @override
    def with_spatial_borders_trimmed(self, borders: Iterable[SpatialBlockBorder]) -> LayoutLayer:
        raise TQECError(f"Cannot trim spatial borders of a {type(self).__name__} instance.")

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, LayoutLayer)
            and self.element_shape == value.element_shape
            and self.layers == value.layers
        )

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    def reschedule_measurements(self) -> None:
        """Re-schedule measurements to be in the same moment."""
        # Collect all plaquettes from all layers
        plaquettes = [
            plaquette
            for layer in self.layers.values()
            if isinstance(layer, PlaquetteLayer)
            for plaquette in layer.plaquettes.collection.values()
        ]
        if not plaquettes:
            return
        # Find the maximum schedule value
        max_schedule = max(plaquette.circuit.schedule.max_schedule for plaquette in plaquettes)
        # Reschedule all plaquettes
        for plaquette in plaquettes:
            plaquette.reschedule_measurements(max_schedule)

    def to_template_and_plaquettes(self) -> tuple[LayoutTemplate, Plaquettes]:
        """Return an equivalent representation of ``self`` with a template and some plaquettes.

        Raises:
            NotImplementedError: if not all layers composing ``self`` are instances
                of :class:`~tqec.compile.blocks.layers.atomic.plaquette.PlaquetteLayer`.

        Returns:
            a tuple ``(template, plaquettes)`` that is ready to be used with
            :meth:`~tqec.compile.generation.generate_circuit` to obtain the quantum
            circuit representing ``self``.

        """
        if not contains_only_plaquette_layers(self.layers):
            raise NotImplementedError(
                f"Found a layer that is not an instance of {PlaquetteLayer.__name__}. "
                "Detector computation is not implemented (yet) for this case."
            )
        cubes: dict[BlockPosition2D, PlaquetteLayer] = {
            pos.to_block_position(): layer
            for pos, layer in self.layers.items()
            if isinstance(pos, LayoutCubePosition2D) and isinstance(layer, PlaquetteLayer)
        }
        template_dict: Final = {pos: layer.template for pos, layer in cubes.items()}
        plaquettes_dict = {pos: layer.plaquettes for pos, layer in cubes.items()}

        # Add plaquettes from each pipe to the plaquette_dict.
        pipes: dict[tuple[BlockPosition2D, BlockPosition2D], PlaquetteLayer] = {
            pos.to_pipe(): layer
            for pos, layer in self.layers.items()
            if isinstance(pos, LayoutPipePosition2D) and isinstance(layer, PlaquetteLayer)
        }
        for (u, v), pipe_layer in pipes.items():
            pipe_direction = Direction3D.from_neighbouring_positions(u.to_3d(), v.to_3d())
            # {u,v}_border: border of the respective node that is touched by the
            # the pipe.
            u_border: TemplateBorder
            v_border: TemplateBorder
            match pipe_direction:
                case Direction3D.X:
                    u_border, v_border = TemplateBorder.RIGHT, TemplateBorder.LEFT
                case Direction3D.Y:
                    u_border, v_border = TemplateBorder.BOTTOM, TemplateBorder.TOP
                case Direction3D.Z:
                    raise TQECError("Should not happen. This is a logical error.")

            # Updating plaquettes in plaquettes_dict
            for pos, (cube_border, pipe_border) in [
                (u, (u_border, v_border)),
                (v, (v_border, u_border)),
            ]:
                plaquette_indices_mapping = pipe_layer.template.get_border_indices(pipe_border).to(
                    template_dict[pos].get_border_indices(cube_border)
                )
                pipe_plaquette_collection = pipe_layer.plaquettes.collection
                plaquettes_dict[pos] = plaquettes_dict[pos].with_updated_plaquettes(
                    {
                        plaquette_indices_mapping[pipe_plaquette_index]: plaquette
                        for pipe_plaquette_index, plaquette in pipe_plaquette_collection.items()
                        # Filtering only plaquette indices that are on the side we are
                        # interested in.
                        if pipe_plaquette_index in plaquette_indices_mapping
                    }
                )

        template = LayoutTemplate(template_dict)
        return template, template.get_global_plaquettes(plaquettes_dict)

    def to_circuit(self, k: int, reschedule_measurements: bool = True) -> ScheduledCircuit:
        """Return the quantum circuit representing the layer.

        Args:
            k: scaling factor.
            reschedule_measurements: whether to reschedule measurements in the generated circuit
                to be in the same moment.
                Since each plaquette may have its own measurement schedule, setting this
                may be necessary for hardware that requires measurements to be synchronous.

        Returns:
            quantum circuit representing the layer.

        """
        if reschedule_measurements:
            self.reschedule_measurements()
        template, plaquettes = self.to_template_and_plaquettes()
        scheduled_circuit = generate_circuit(template, k, plaquettes)
        # Shift the qubits of the returned scheduled circuit
        mincube, _ = self.bounds
        eshape = self.element_shape.to_shape_2d(k)
        # See: https://github.com/tqec/tqec/issues/525
        # This is a temporary fix to the above issue, we may need a utility function
        # to calculate shift to avoid similar issues in the future.
        shift = Shift2D(mincube.x * (eshape.x - 1), mincube.y * (eshape.y - 1))
        shifted_circuit = scheduled_circuit.map_to_qubits(lambda q: q + shift)
        return shifted_circuit

    @property
    @override
    def scalable_num_moments(self) -> LinearFunction:
        return LinearFunction.unambiguous_max_on_positives(
            layer.scalable_num_moments for layer in self.layers.values()
        )

    @property
    def qubit_bounds(self) -> tuple[PhysicalQubitScalable2D, PhysicalQubitScalable2D]:
        """Return the top-left and bottom-right qubits representing the bounding box of ``self``.

        Returns:
            the ``(top_left, bottom_right)`` qubits.

        """
        tlb, brb = self.bounds
        eshape = self.element_shape
        increments = PhysicalQubitScalable2D(eshape.x, eshape.y) - (1, 1)
        tlq = PhysicalQubitScalable2D(tlb.x * increments.x, tlb.y * increments.y)
        brq = PhysicalQubitScalable2D((brb.x + 1) * increments.x, (brb.y + 1) * increments.y)
        # Note: for the moment, plaquette origin is defined as the CENTER of the
        # plaquette, which lead the above computations to be off-by-1. Correct that
        # before returning.
        shift = (-1, -1)
        return tlq + shift, brq + shift



"""Defines the data structures to represent the logical computations.

This module provides high-level abstractions to represent the fault-tolerant logical computations
protected by surface code.

"""



"""Block graph that represent a CNOT gate."""

from tqec.computation.block_graph import BlockGraph
from tqec.computation.cube import ZXCube
from tqec.utils.enums import Basis
from tqec.utils.position import Position3D


def cnot(observable_basis: Basis | None = None) -> BlockGraph:
    """Create a block graph for the logical CNOT gate.

    Args:
        observable_basis: The observable basis that the block graph can support.
            If None, the four ports of the block graph will be left open.
            Otherwise, the ports will be filled with the cubes that have the
            initializations and measurements in the given observable basis.

    Returns:
        A :py:class:`~tqec.computation.block_graph.BlockGraph` instance representing
        the logical CNOT gate.

    """
    g = BlockGraph("Logical CNOT")
    nodes = [
        (Position3D(0, 0, 0), "P", "In_Control"),
        (Position3D(0, 0, 1), "ZXX", ""),
        (Position3D(0, 0, 2), "ZXZ", ""),
        (Position3D(0, 0, 3), "P", "Out_Control"),
        (Position3D(0, 1, 1), "ZXX", ""),
        (Position3D(0, 1, 2), "ZXZ", ""),
        (Position3D(1, 1, 0), "P", "In_Target"),
        (Position3D(1, 1, 1), "ZXZ", ""),
        (Position3D(1, 1, 2), "ZXZ", ""),
        (Position3D(1, 1, 3), "P", "Out_Target"),
    ]
    for pos, kind, label in nodes:
        g.add_cube(pos, kind, label)

    pipes = [(0, 1), (1, 2), (2, 3), (1, 4), (4, 5), (5, 8), (6, 7), (7, 8), (8, 9)]

    for p0, p1 in pipes:
        g.add_pipe(nodes[p0][0], nodes[p1][0])

    if observable_basis == Basis.Z:
        g.fill_ports(ZXCube.from_str("ZXZ"))
    elif observable_basis == Basis.X:
        g.fill_ports(ZXCube.from_str("ZXX"))
    return g



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any

from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap
from tqec.plaquette.enums import PlaquetteSide
from tqec.utils.enums import Orientation


@dataclass(frozen=True)
class PlaquetteQubits:
    data_qubits: list[GridQubit]
    syndrome_qubits: list[GridQubit]

    def __iter__(self) -> Iterator[GridQubit]:
        yield from self.data_qubits
        yield from self.syndrome_qubits

    @property
    def all_qubits(self) -> list[GridQubit]:
        """Return all the qubits represented by ``self``."""
        return list(self)

    def get_edge_qubits(
        self,
        orientation: Orientation = Orientation.HORIZONTAL,
    ) -> list[GridQubit]:
        """Return the data qubits on the edge of the plaquette.

        By convention, the edge is the one with the highest index in the relevant axis.

        Args:
            orientation (TemplateOrientation, optional): Whether to use horizontal or
                vertical orientation as the axis. Defaults to horizontal.

        Returns:
            The qubits on the edge of the plaquette.

        """

        def _get_relevant_value(qubit: GridQubit) -> int:
            return qubit.y if orientation == Orientation.HORIZONTAL else qubit.x

        max_index = max(_get_relevant_value(q) for q in self.data_qubits)
        return [qubit for qubit in self.data_qubits if (_get_relevant_value(qubit) == max_index)]

    def get_qubits_on_side(self, side: PlaquetteSide) -> list[GridQubit]:
        """Return the qubits one the provided side of the instance.

        A qubit is on the left-side if there is no other qubit in the instance
        with a strictly lower x-coordinate value. Similarly, a qubit is on the
        right-side if there is no other qubit in the instance with a strictly
        greater x-coordinate value. Up and down are about the y-coordinate.

        Args:
            side: the side to find qubits on.

        Returns:
            The qubits on the edge of the plaquette.

        """
        if side == PlaquetteSide.LEFT:
            min_x = min(q.x for q in self)
            return [q for q in self if q.x == min_x]
        elif side == PlaquetteSide.RIGHT:
            max_x = max(q.x for q in self)
            return [q for q in self if q.x == max_x]
        elif side == PlaquetteSide.UP:
            min_y = min(q.y for q in self)
            return [q for q in self if q.y == min_y]
        else:  # if orientation == PlaquetteSide.DOWN:
            max_y = max(q.y for q in self)
            return [q for q in self if q.y == max_y]

    def __hash__(self) -> int:
        return hash((tuple(self.syndrome_qubits), tuple(self.data_qubits)))  # pragma: no cover

    def __eq__(self, rhs: object) -> bool:
        return (
            isinstance(rhs, PlaquetteQubits)
            and self.data_qubits == rhs.data_qubits
            and self.syndrome_qubits == rhs.syndrome_qubits
        )

    @property
    def data_qubits_indices(self) -> Iterator[int]:
        """Return an iterator over the index of each data-qubit."""
        yield from range(len(self.data_qubits))

    @property
    def syndrome_qubits_indices(self) -> Iterator[int]:
        """Return an iterator over the index of each syndrome-qubit."""
        yield from (i + len(self.data_qubits) for i in range(len(self.syndrome_qubits)))

    @property
    def data_qubits_with_indices(self) -> Iterator[tuple[int, GridQubit]]:
        """Return an iterator over each data-qubit and their indices."""
        yield from ((i, q) for i, q in zip(self.data_qubits_indices, self.data_qubits))

    @property
    def syndrome_qubits_with_indices(self) -> Iterator[tuple[int, GridQubit]]:
        """Return an iterator over each syndrome-qubit and their indices."""
        yield from ((i, q) for i, q in zip(self.syndrome_qubits_indices, self.syndrome_qubits))

    @property
    def qubit_map(self) -> QubitMap:
        """Return the qubit map representing ``self``."""
        return QubitMap(
            dict(self.data_qubits_with_indices) | dict(self.syndrome_qubits_with_indices)
        )

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the plaquette qubits.

        Returns:
            a dictionary with the keys ``data_qubits`` and ``syndrome_qubits`` and
            their corresponding values.

        """
        return {
            "data_qubits": [q.to_dict() for q in self.data_qubits],
            "syndrome_qubits": [q.to_dict() for q in self.syndrome_qubits],
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> PlaquetteQubits:
        """Return a plaquette qubits from its dictionary representation.

        Args:
            data: dictionary with the keys ``data_qubits`` and
                ``syndrome_qubits``.

        Returns:
            a new instance of :class:`PlaquetteQubits` with the provided
            ``data_qubits`` and ``syndrome_qubits``.

        """
        data_qubits = [GridQubit.from_dict(q) for q in data["data_qubits"]]
        syndrome_qubits = [GridQubit.from_dict(q) for q in data["syndrome_qubits"]]
        return PlaquetteQubits(data_qubits, syndrome_qubits)


class SquarePlaquetteQubits(PlaquetteQubits):
    def __init__(self) -> None:
        """Represent the qubits used by a regular square plaquette."""
        super().__init__(
            # Order is important here! Top-left, top-right, bottom-left,
            # bottom-right.
            [GridQubit(-1, -1), GridQubit(1, -1), GridQubit(-1, 1), GridQubit(1, 1)],
            [GridQubit(0, 0)],
        )



"""Introduces :class:`.PlaquetteCompiler`, the class used to implement a plaquette compiler."""

from collections.abc import Callable, Iterable
from typing import Final

from tqec.plaquette.compilation.passes.base import CompilationPass
from tqec.plaquette.plaquette import Plaquette
from tqec.utils.instructions import (
    MEASUREMENT_INSTRUCTION_NAMES,
    RESET_INSTRUCTION_NAMES,
)


class PlaquetteCompiler:
    def __init__(
        self,
        name: str,
        passes: Iterable[CompilationPass],
        mergeable_instructions_modifier: Callable[[frozenset[str]], frozenset[str]],
    ):
        """Wrap a list of :class:`.CompilationPass` instances."""
        self._name = name
        self._passes = passes
        self._mergeable_instructions_modifier = mergeable_instructions_modifier

    def compile(self, plaquette: Plaquette) -> Plaquette:
        """Apply all the stored compilation passes in order and return the resulting plaquette.

        Args:
            plaquette: plaquette to compile.

        Returns:
            a :class:`~tqec.plaquette.plaquette.Plaquette` instance that has been
            compiled with the compilation passes stored in ``self`` and with a
            modified name.

        """
        circuit = plaquette.circuit
        for compilation_pass in self._passes:
            circuit = compilation_pass.run(circuit)
        return Plaquette(
            f"{self._name}({plaquette.name})",
            plaquette.qubits,
            circuit,
            self._mergeable_instructions_modifier(plaquette.mergeable_instructions),
            plaquette.debug_information,
        )


IdentityPlaquetteCompiler: Final[PlaquetteCompiler] = PlaquetteCompiler(
    "ID", [], lambda x: x | MEASUREMENT_INSTRUCTION_NAMES | RESET_INSTRUCTION_NAMES
)



"""Defines several helper methods to facilitate visualising QEC circuit simulation."""

from .inset import add_inset_axes3d as add_inset_axes3d
from .inset import plot_observable_as_inset as plot_observable_as_inset



"""Provide a flexible representation to define blocks.

The main data-structure provided by this module is
:class:`~tqec.compile.blocks.block.Block`. It is able to represent in a flexible
manner anything that looks like a block in a topological computation represented
using SketchUp. In particular, that data-structure can be used to represent both
cubes and pipes.

In a topological computation, :class:`~tqec.compile.blocks.block.Block`
instances will happen in parallel, and ``tqec`` needs to account for that:
operations happening in parallel should be encoded in the same moment (i.e.,
between the same two ``TICK`` instructions) in the resulting ``.stim`` file.

Moreover, blocks might have different "schedules". For example, the regular
memory block follow the schedule:

1. initialisation layer,
2. repeat [memory layer],
3. measurement layer,

whereas a spatial pipe in the ``Y`` axis needs to alternate plaquettes in its
repeated layer, leading to a schedule that is:

1. initialisation layer,
2. repeat [memory layer 1 alternated with memory layer 2],
3. measurement layer.

All these restrictions are handled by representing
:class:`~tqec.compile.blocks.block.Block` instances with
:class:`~tqec.compile.blocks.layers.atomic.base.BaseLayer` and
:class:`~tqec.compile.blocks.layers.composed.base.BaseComposedLayer` instances.
See :mod:`tqec.compile.blocks.layers` for more details.

"""



"""Defines functions to modify or merge :class:`.ScheduledCircuit` instances.

This module implement a few central functions for the :mod:`tqec` library:

- :func:`remove_duplicate_instructions` to remove some instructions appearing
  twice in a single moment (most of the time due to data qubit
  reset/measurements that are defined by each plaquette, even on qubits shared
  with other plaquettes, leading to duplicates).
- :func:`merge_scheduled_circuits` that merge several
  :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit` instances into one.
- :func:`relabel_circuits_qubit_indices` to prepare several
  :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit` before merging them.
  This function is called internally by :func:`merge_scheduled_circuits` but
  might be useful at other places and so is kept public.

"""

from __future__ import annotations

import functools
import itertools
import operator
import warnings
from collections.abc import Iterable, Sequence

import stim

from tqec.circuit.moment import Moment
from tqec.circuit.qubit import GridQubit
from tqec.circuit.qubit_map import QubitMap
from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.circuit.schedule.schedule import Schedule
from tqec.utils.exceptions import TQECError, TQECWarning


class _ScheduledCircuits:
    def __init__(self, circuits: list[ScheduledCircuit], global_qubit_map: QubitMap) -> None:
        """Represent a collection of :class:`.ScheduledCircuit` instances.

        This class aims at providing accessors for several compatible instances
        of :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit`. It allows
        to iterate on gates globally, for all the managed instances of
        :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit`, and implement
        a few other accessor methods to help with the task of merging multiple
        :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit` together.

        Args:
            circuits: the instances that should be managed. Note that the
                instances provided here have to be "compatible" with each
                other.
            global_qubit_map: a unique qubit map that can be used to map qubits
                to indices for all the provided ``circuits``.

        """
        # We might need to remap qubits to avoid index collision on several
        # circuits.
        self._circuits = circuits
        self._global_qubit_map = global_qubit_map
        self._iterators = [circuit.scheduled_moments for circuit in self._circuits]
        self._current_moments = [next(it, None) for it in self._iterators]

    def has_pending_moment(self) -> bool:
        """Check if any of the managed instances has a pending moment.

        Any moment that has not been collected by using collect_moment is considered to be pending.

        Any moment that has not been collected by using ``collect_moment`` is considered to be
        pending.

        """
        return any(self._has_pending_moment(i) for i in range(len(self._circuits)))

    def _has_pending_moment(self, index: int) -> bool:
        """Check if the managed instance at the given index has a pending operation."""
        return self._current_moments[index] is not None

    def _peek_scheduled_moment(self, index: int) -> tuple[int, Moment]:
        """Recover **without collecting** pending operations for the instance at the given index."""
        ret = self._current_moments[index]
        assert ret is not None
        return ret

    def _pop_scheduled_moment(self, index: int) -> tuple[int, Moment]:
        """Recover and mark as collected the pending moment for the instance at the given index.

        Raises:
            AssertionError: ``if not self.has_pending_operation(index)``.

        """
        ret = self._current_moments[index]
        if ret is None:
            raise TQECError(
                "Trying to pop a Moment instance from a ScheduledCircuit with "
                "all its moments already collected."
            )
        self._current_moments[index] = next(self._iterators[index], None)
        return ret

    @property
    def number_of_circuits(self) -> int:
        return len(self._circuits)

    def collect_moments_at_minimum_schedule(self) -> tuple[int, list[Moment]]:
        """Collect all the moments that can be collected.

        This method collects and returns a list of all the moments that should
        be scheduled next.

        Returns:
            a list of :class:`~tqec.circuit.moment.Moment` instances that should
            be added next to the QEC circuit.

        """
        assert self.has_pending_moment()
        circuit_indices_organised_by_schedule: dict[int, list[int]] = dict()
        for circuit_index in range(self.number_of_circuits):
            if not self._has_pending_moment(circuit_index):
                continue
            schedule, _ = self._peek_scheduled_moment(circuit_index)
            circuit_indices_organised_by_schedule.setdefault(schedule, list()).append(circuit_index)

        minimum_schedule = min(circuit_indices_organised_by_schedule.keys())
        moments_to_return: list[Moment] = list()
        for circuit_index in circuit_indices_organised_by_schedule[minimum_schedule]:
            _, moment = self._pop_scheduled_moment(circuit_index)
            moments_to_return.append(moment)
        return minimum_schedule, moments_to_return

    @property
    def q2i(self) -> dict[GridQubit, int]:
        return self._global_qubit_map.q2i


def _sort_target_groups(
    targets: Iterable[list[stim.GateTarget]],
) -> list[list[stim.GateTarget]]:
    def _sort_key(target_group: list[stim.GateTarget]) -> tuple[int, ...]:
        return tuple(t.value for t in target_group)

    return sorted(targets, key=_sort_key)


def remove_duplicate_instructions(
    instructions: list[stim.CircuitInstruction],
    mergeable_instruction_names: frozenset[str],
) -> list[stim.CircuitInstruction]:
    """Remove all the duplicate instructions from the given list.

    Note:
        This function guarantees the following post-conditions on the returned
        results:

        - Instructions with a name that is not in ``mergeable_instruction_names``
          are returned at the front of the returned list, in the same relative
          ordering as provided in ``instructions`` input.
        - Instructions with a name that is in ``mergeable_instruction_names``
          will be returned after all the instructions with a name that does not
          appear in ``mergeable_instruction_names``. The order in which these
          instructions are returned is not guaranteed and can change between
          executions.

    Warning:
        this function **does not keep instruction ordering**. It is intended to
        be used with input ``instructions`` that, once de-duplication has been
        applied, form a valid moment, which means that each instruction can be
        executed in parallel, and so their order in the returned list does not
        matter.

        If that is not your case, take extra care to the output of this
        function as it will likely introduce hard-to-debug issues. To prevent
        such potential misuse, this function checks for such cases and outputs
        a warning if it happens.

    Returns:
        a list containing a copy of the ``stim.CircuitInstruction`` instances
        from the given instructions but without any duplicate.

    """
    # Separate mergeable operations from non-mergeable ones.
    mergeable_operations: dict[tuple[str, tuple[float, ...]], set[tuple[stim.GateTarget, ...]]] = {}
    final_operations: list[stim.CircuitInstruction] = list()
    for inst in instructions:
        if inst.name in mergeable_instruction_names:
            # Mergeable operations are automatically merged thanks to
            # the use of a set here.
            mergeable_operations.setdefault(
                (inst.name, tuple(inst.gate_args_copy())), set()
            ).update(tuple(group) for group in inst.target_groups())
        else:
            final_operations.append(inst)
    # Add the merged operations into the final ones
    final_operations.extend(
        stim.CircuitInstruction(
            name,
            functools.reduce(operator.iadd, _sort_target_groups([list(t) for t in targets]), []),
            args,
        )
        for (name, args), targets in mergeable_operations.items()
    )
    # Warn if the output instructions do not form a valid moment, as this is
    # likely a misuse of this function.
    circuit = stim.Circuit()
    for instr in final_operations:
        circuit.append(instr)
    try:
        Moment.check_is_valid_moment(circuit)
    except TQECError as e:
        warnings.warn(
            "The instructions obtained at the end of the "
            "`remove_duplicate_instructions` function do not form a valid "
            "moment. You are likely misusing the function. Final instructions "
            "obtained and gathered into a single stim.Circuit: "
            f"\n{circuit}\nReason:\n{e}",
            TQECWarning,
        )
    return final_operations


def merge_instructions(
    instructions: list[stim.CircuitInstruction],
) -> list[stim.CircuitInstruction]:
    """Merge instructions with the same name and arguments.

    Returns:
        a list containing a copy of the ``stim.CircuitInstruction`` instances
        from the given instructions but merged.

    """
    instructions_merger: dict[tuple[str, tuple[float, ...]], list[list[stim.GateTarget]]] = {}
    for instruction in instructions:
        args = tuple(instruction.gate_args_copy())
        instructions_merger.setdefault((instruction.name, args), []).extend(
            instruction.target_groups()
        )
    return [
        stim.CircuitInstruction(name, functools.reduce(operator.iadd, targets, []), args)
        for (name, args), targets in instructions_merger.items()
    ]


def merge_scheduled_circuits(
    circuits: list[ScheduledCircuit],
    global_qubit_map: QubitMap,
    mergeable_instructions: Iterable[str] = (),
) -> ScheduledCircuit:
    """Merge several :class:`.ScheduledCircuit` instances into one instance.

    This function takes several **compatible** scheduled circuits as input and
    merge them, respecting their schedules, into a unique
    :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit` instance that will
    then be returned to the caller.

    The provided circuits should be compatible between each other. Compatible
    circuits are circuits that can all be described with a unique global qubit
    map. In other words, if two circuits from the list of compatible circuits
    use the same qubit index, that should mean that they use the same qubit.
    You can obtain compatible circuits by using
    :func:`relabel_circuits_qubit_indices`.

    Args:
        circuits: **compatible** circuits to merge.
        global_qubit_map: global qubit map for all the provided ``circuits``.
        mergeable_instructions: a list of instruction names that are considered
            mergeable. Duplicate instructions with a name in this list will be
            merged into a single instruction.

    Returns:
        a circuit representing the merged scheduled circuits given as input.

    """
    scheduled_circuits = _ScheduledCircuits(circuits, global_qubit_map)

    all_moments: list[Moment] = []
    all_schedules = Schedule()
    global_i2q = QubitMap({i: q for q, i in scheduled_circuits.q2i.items()})

    while scheduled_circuits.has_pending_moment():
        schedule, moments = scheduled_circuits.collect_moments_at_minimum_schedule()
        # Flatten the moments into a list of operations to perform some modifications
        instructions: list[stim.CircuitInstruction] = functools.reduce(
            operator.iadd, (list(moment.instructions) for moment in moments), []
        )
        # Avoid duplicated operations. Any operation that have the Plaquette.get_mergeable_tag() tag
        # is considered mergeable, and can be removed if another operation in the list
        # is considered equal (and has the mergeable tag).
        deduplicated_instructions = remove_duplicate_instructions(
            instructions,
            mergeable_instruction_names=frozenset(mergeable_instructions),
        )
        merged_instructions = merge_instructions(deduplicated_instructions)
        circuit = stim.Circuit()
        for inst in merged_instructions:
            circuit.append(
                inst.name,
                functools.reduce(operator.iadd, _sort_target_groups(inst.target_groups()), []),
                inst.gate_args_copy(),
            )
        all_moments.append(Moment(circuit))
        all_schedules.append(schedule)

    return ScheduledCircuit(all_moments, all_schedules, global_i2q, _avoid_checks=True)


def relabel_circuits_qubit_indices(
    circuits: Sequence[ScheduledCircuit],
) -> tuple[list[ScheduledCircuit], QubitMap]:
    """Relabel the qubit indices of the provided circuits to avoid collision.

    When several :class:`~tqec.circuit.schedule.circuit.ScheduledCircuit` are
    constructed without a global knowledge of all the qubits, qubit indices used
    by each instance likely overlap. This is an issue when we try to merge such
    circuits because one index might represent a different qubit depending on
    the circuit it is used in.

    This function takes a sequence of circuits and relabel their qubits to avoid
    such collisions.

    Warning:
        all the qubit targets used in each of the provided circuits should have
        a corresponding entry in the circuit qubit map for this function to work
        correctly. If that is not the case, a ``KeyError`` will be raised.

    Raises:
        KeyError: if any of the provided circuit contains a qubit target that is
            not present in its qubit map.

    Args:
        circuits: circuit instances to remap. This parameter is not mutated by
            this function and is only used in read-only mode.

    Returns:
        the same circuits with updated qubit indices as well as the global qubit
        indices map that has been used. Qubits in the returned global qubit map
        are assigned to an index such that:

        1. the sequence of indices is ``range(0, len(qubit_map))``.
        2. qubits are assigned indices in sorted order.

    """
    # First, get a global qubit index map.
    # Using itertools to avoid the edge case `len(circuits) == 0`
    needed_qubits = frozenset(itertools.chain.from_iterable([c.qubits for c in circuits]))
    global_qubit_map = QubitMap.from_qubits(sorted(needed_qubits))
    global_q2i = global_qubit_map.q2i
    # Then, get the remapped circuits. Note that map_qubit_indices should
    # have approximately the same runtime cost whatever the value of inplace
    # so we ask for a new instance to avoid keeping a reference to the given
    # circuits.
    relabeled_circuits: list[ScheduledCircuit] = []
    for circuit in circuits:
        local_indices_to_global_indices = {
            local_index: global_q2i[q] for local_index, q in circuit.qubit_map.items()
        }
        relabeled_circuits.append(
            circuit.map_qubit_indices(local_indices_to_global_indices, inplace=False)
        )
    return relabeled_circuits, global_qubit_map



from collections.abc import Sequence

import stim
import svg

from tqec.circuit.qubit import GridQubit
from tqec.interop.color import TQECColor
from tqec.visualisation.exception import TQECDrawingError


def _get_error_cross_svg(
    size: float = 1, stroke_color: str = "red", stroke_width_multiplier: float = 0.2
) -> svg.Element:
    """Return an SVG element representing an error."""
    return svg.Path(
        d=[
            svg.M(-size / 2, -size / 2),
            svg.L(size / 2, size / 2),
            svg.M(-size / 2, size / 2),
            svg.L(size / 2, -size / 2),
        ],
        stroke=stroke_color,
        stroke_width=size * stroke_width_multiplier,
    )


def _get_coordinates(
    targets: list[stim.GateTargetWithCoords],
) -> tuple[float, float, float | None, float | None]:
    match len(targets):
        case 1:
            qx, qy = targets[0].coords
            return qx, qy, None, None
        case 2:
            qx, qy = targets[0].coords
            qx2, qy2 = targets[1].coords
            return qx, qy, qx2, qy2
        case 0:
            raise TQECDrawingError("Could not extract coordinates from an empty list.")
        case _:
            raise TQECDrawingError(
                f"Could not extract coordinates from a list with {len(targets)}."
            )


def get_errors_svg(
    errors: Sequence[stim.ExplainedError],
    top_left_qubit: GridQubit,
    plaquette_width: float,
    plaquette_height: float,
    size: float = 1,
    add_detectors_from_errors: bool = True,
) -> svg.G:
    """Return an SVG element with the provided ``errors`` drawn and a transparent background.

    Args:
        errors: a sequence of errors to plot. It is often desirable to filter
            errors according to the moment they appear in.
        top_left_qubit: coordinates of the qubit at the very top-left of the
            visualisation canvas. Used to correctly offset qubit values from the
            provided ``errors``.
        plaquette_width: width (in SVG dimensions) of a regular square plaquette.
        plaquette_height: height (in SVG dimensions) of a regular square plaquette.
        size: size (in SVG dimensions) of the crosses used to mark errors.
        add_detectors_from_errors: add a visualisation of detectors from the provided errors.

    Returns:
        a SVG element containing as many sub-elements as there are errors in the
        provided ``errors`` list, each representing one of the provided errors.

    Raises:
        TQECDrawingError: if at least one of the provided ``errors`` has both an
            empty ``flipped_measurement`` list AND an empty ``flipped_pauli_product``
            list.

    """
    # Default value for the moment
    stroke_width_multiplier = 0.2
    width_between_qubits: float = plaquette_width / 2
    height_between_qubits: float = plaquette_height / 2
    stroke_width: float = size * stroke_width_multiplier
    layer: list[svg.Element] = []
    # Start by adding the detectors
    detectors: dict[int, stim.DemTargetWithCoords] = {}
    if add_detectors_from_errors:
        # List the detectors that need to be drawn
        for error in errors:
            for target in error.dem_error_terms:
                if target.dem_target.is_relative_detector_id():
                    detectors[target.dem_target.val] = target
        # Draw each of them once
        for detector in detectors.values():
            x, y, t, *_ = detector.coords
            layer.append(
                svg.G(
                    elements=[
                        svg.Circle(
                            cx=0,
                            cy=0,
                            r=0.15,
                            stroke="black",
                            fill="transparent",
                            stroke_width=0.01,
                        ),
                        svg.Text(
                            x=0,
                            y=0,
                            text=f"D{detector.dem_target.val} ({int(t)})",
                            fill="black",
                            font_size=0.08,
                            text_anchor="middle",
                            dominant_baseline="middle",
                        ),
                    ],
                    transform=[
                        svg.Translate(
                            (x - top_left_qubit.x) * width_between_qubits,
                            (y - top_left_qubit.y) * height_between_qubits,
                        )
                    ],
                )
            )
    for error in errors:
        # We take the first error location. All the error locations in error.circuit_error_locations
        # are valid and could be picked, but we need to take one here.
        location = error.circuit_error_locations[0]
        # Get the x and y coordinates in qubit-coordinates.
        # For 2-qubit errors, we get also the second qubit x and y coordinates.
        flipped_measurement = location.flipped_measurement
        flipped_pauli_product = location.flipped_pauli_product
        if flipped_pauli_product:
            qx, qy, qx2, qy2 = _get_coordinates(flipped_pauli_product)
            basis = flipped_pauli_product[0].gate_target.pauli_type
        elif flipped_measurement:
            qx, qy, qx2, qy2 = _get_coordinates(flipped_measurement.observable)
            basis = flipped_measurement.observable[0].gate_target.pauli_type
        else:
            raise TQECDrawingError("Could not draw the following error:\n" + str(error))
        color = TQECColor(f"{basis}_CORRELATION").rgba.to_hex()
        # Make the coordinates relative to the top-left qubit.
        qx -= top_left_qubit.x
        qy -= top_left_qubit.y
        # Plot the cross for a single-qubit error.
        tick_text_svg = svg.Text(
            x=0,
            y=-size,
            text=str(location.tick_offset),
            fill=color,
            font_size=0.3,
            font_weight="bold",
            text_anchor="middle",
            dominant_baseline="auto",
        )
        cross_svg = _get_error_cross_svg(size, color, stroke_width_multiplier)
        error_svg: list[svg.Element] = [
            svg.G(
                elements=[cross_svg, tick_text_svg],
                transform=[svg.Translate(qx * width_between_qubits, qy * height_between_qubits)],
            ),
        ]
        detector_arrow_source = qx * width_between_qubits, qy * height_between_qubits
        # Plot the additional SVG lines if we have a 2-qubit error.
        if qx2 is not None and qy2 is not None:
            qx2 -= top_left_qubit.x
            qy2 -= top_left_qubit.y
            error_svg.append(
                svg.G(
                    elements=[cross_svg],
                    transform=[
                        svg.Translate(qx2 * width_between_qubits, qy2 * height_between_qubits)
                    ],
                )
            )
            error_svg.append(
                svg.Line(
                    x1=qx * width_between_qubits,
                    x2=qx2 * width_between_qubits,
                    y1=qy * height_between_qubits,
                    y2=qy2 * height_between_qubits,
                    stroke=color,
                    stroke_width=stroke_width,
                )
            )
            detector_arrow_source = (
                (qx + qx2) * width_between_qubits / 2,
                (qy + qy2) * height_between_qubits / 2,
            )
        # Plot lines to link detectors and errors.
        if add_detectors_from_errors:
            for dem_error_term in error.dem_error_terms:
                if not dem_error_term.dem_target.is_relative_detector_id():
                    continue
                x, y, t, *_ = dem_error_term.coords
                error_svg.append(
                    svg.Line(
                        x1=detector_arrow_source[0],
                        y1=detector_arrow_source[1],
                        x2=(x - top_left_qubit.x) * width_between_qubits,
                        y2=(y - top_left_qubit.y) * height_between_qubits,
                        stroke="black",
                        stroke_width=0.01,
                    )
                )
        layer.append(svg.G(elements=error_svg))
    return svg.G(elements=layer)
